/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/**
 * Defines {@link baja.RelTime}.
 * @module baja/obj/RelTime
 */
define(["bajaScript/sys",
        "bajaScript/baja/obj/Simple",
        "bajaScript/baja/obj/dateTimeUtil",
        "bajaScript/baja/obj/objUtil"], 
        function (baja, Simple, dateTimeUtil, objUtil) {
 "use strict";
 
 var subclass = baja.subclass,
     callSuper = baja.callSuper,
     objectify = baja.objectify,
     strictArg = baja.strictArg,
     bajaDef = baja.def,
     
     cacheDecode = objUtil.cacheDecode,
     cacheEncode = objUtil.cacheEncode,
     
     MILLIS_IN_SECOND = dateTimeUtil.MILLIS_IN_SECOND,
     MILLIS_IN_MINUTE = dateTimeUtil.MILLIS_IN_MINUTE,
     MILLIS_IN_HOUR = dateTimeUtil.MILLIS_IN_HOUR,
     MILLIS_IN_DAY = dateTimeUtil.MILLIS_IN_DAY;
 
  /**
   * Represents a `baja:RelTime` in BajaScript.
   * 
   * `RelTime` is a `Simple` type for managing a relative amount of time. 
   * 
   * When creating a `Simple`, always use the `make()` method instead of 
   * creating a new Object.
   *
   * @class
   * @alias baja.RelTime
   * @extends baja.Simple
   */
  var RelTime = function RelTime(ms) {
    callSuper(RelTime, this, arguments); 
    this.$ms = parseInt(ms, 10);
  };
  
  subclass(RelTime, Simple);
  
  /**
   * Make a `RelTime`.
   * 
   * @param {Object|Number} [obj] the Object Literal or the number of milliseconds.
   * @param {Number} [obj.days] the number of days.
   * @param {Number} [obj.hours] the number of hours.
   * @param {Number} [obj.minutes] the number of minutes.
   * @param {Number} [obj.seconds] the number of seconds.
   * @param {Number} [obj.ms] the number of milliseconds.
   * @returns {baja.RelTime}
   * 
   * @example
   *  //This method can take a number of milliseconds or an Object Literal as 
   *  //method's argument...
   *   var rt1 = baja.RelTime.make(1000); // One second
   *
   *   // ...or we can specify an Object Literal for more arguments...
   *
   *   // Create a RelTime with 2 days + 2 hours + 2 minutes + 2 seconds + 2 milliseconds...
   *   var rt2 = baja.RelTime.make({
   *     days: 2,
   *     hours: 2,
   *     minutes: 2,
   *     seconds: 2,
   *     ms: 2
   *   });
   */
  RelTime.make = function (obj) {
    var ms = 0;
    
    if (typeof obj === "number") {
      ms = obj;
    }
    else {
      obj = objectify(obj, "ms");
      ms = bajaDef(obj.ms, 0);
      
      strictArg(ms, Number);
      
      if (typeof obj.days === "number") {
        ms += MILLIS_IN_DAY * obj.days;
      }
      if (typeof obj.hours === "number") {
        ms += MILLIS_IN_HOUR * obj.hours;
      }
      if (typeof obj.minutes === "number") {
        ms += MILLIS_IN_MINUTE * obj.minutes;
      }
      if (typeof obj.seconds === "number") {
        ms += MILLIS_IN_SECOND * obj.seconds;
      }
    }
        
    if (ms === 0) {
      return RelTime.DEFAULT;
    }
    return new RelTime(ms);
  };
  
  /**
   * Make a `RelTime`.
   * 
   * @param {Object|Number} [obj] the Object Literal or the number of milliseconds.
   * @param {Number} [obj.days] the number of days.
   * @param {Number} [obj.hours] the number of hours.
   * @param {Number} [obj.minutes] the number of minutes.
   * @param {Number} [obj.seconds] the number of seconds.
   * @param {Number} [obj.ms] the number of milliseconds.
   * @returns {baja.RelTime}
   * 
   * @example
   * //This method can take a number of milliseconds of an Object Literal with 
   * //the method's argument...
   *   var rt1 = baja.$("baja:RelTime").make(1000); // One second
   *
   *   // ...or we can specify an Object Literal for more arguments...
   *
   *   // Create a RelTime with 2 days + 2 hours + 2 minutes + 2 seconds + 2 milliseconds...
   *   var rt2 = baja.$("baja:RelTime").make({
   *     days: 2,
   *     hours: 2,
   *     minutes: 2,
   *     seconds: 2,
   *     ms: 2
   *   });
   */
  RelTime.prototype.make = function (obj) {
    return RelTime.make.apply(RelTime, arguments);
  };
  
  /**
   * Decode a `RelTime` from a `String`.
   *
   * @method
   * @param {String} str
   * @returns {baja.RelTime}
   */  
  RelTime.prototype.decodeFromString = cacheDecode(function (str) {
    // Parse number
    var n = Number(str);
    
    // If still not a number then throw an error
    if (isNaN(n)) {
      throw new Error("Unable to create RelTime: " + str);
    }
  
    return RelTime.make(n);
  });
  
  /**
   * Encode the `RelTime` to a `String`.
   *
   * @method
   * @returns {String}
   */ 
  RelTime.prototype.encodeToString = cacheEncode(function () {
    return this.$ms.toString();
  });
  
  /**
   * Default `RelTime` instance.
   * @type {baja.RelTime}
   */   
  RelTime.DEFAULT = new RelTime(0);
        
  /**
   * Milliseconds in a second.
   * @type {Number}
   */   
  RelTime.MILLIS_IN_SECOND = MILLIS_IN_SECOND;
  
  /**
   * Milliseconds in a minute.
   * @type {Number}
   */  
  RelTime.MILLIS_IN_MINUTE = MILLIS_IN_MINUTE;
  
  /**
   * Milliseconds in an hour.
   * @type {Number}
   */  
  RelTime.MILLIS_IN_HOUR = MILLIS_IN_HOUR;
  
  /**
   * Milliseconds in a day.
   * @type {Number}
   */  
  RelTime.MILLIS_IN_DAY = MILLIS_IN_DAY;
  
  /**
   * `RelTime` instance for a second.
   * @type {baja.RelTime}
   */  
  RelTime.SECOND = RelTime.make(MILLIS_IN_SECOND);
  
  /**
   * `RelTime` instance for a minute.
   * @type {baja.RelTime}
   */
  RelTime.MINUTE = RelTime.make(MILLIS_IN_MINUTE);
  
  /**
   * `RelTime` instance for an hour.
   * @type {baja.RelTime}
   */
  RelTime.HOUR = RelTime.make(MILLIS_IN_HOUR);
  
  /**
   * `RelTime` instance for a day.
   * @type {baja.RelTime}
   */
  RelTime.DAY = RelTime.make(MILLIS_IN_DAY);
  
  /**
   * Equality test.
   *
   * @param obj
   * @returns {Boolean}
   */
  RelTime.prototype.equals = function (obj) {
    return objUtil.valueOfEquals(this, obj);
  };
  
  /**
   * Return the data type symbol.
   * 
   * @returns {String}
   */   
  RelTime.prototype.getDataTypeSymbol = function () {
    return "r";
  };
  
  /**
   * Return number of milliseconds.
   *
   * @returns {Number}
   */    
  RelTime.prototype.valueOf = function () {
    return this.$ms;
  };

  /**
   * Return a `String` representation of a `RelTime`.
   *
   * @param {Object} [cx] the context.
   * @param {Boolean} [cx.showMillis=true] set to false to hide milliseconds.
   * @param {Boolean} [cx.showSeconds=true] set to false to hide seconds. If
   * showSeconds is false then milliseconds will not be shown.
   * @param {Boolean} [cx.showMinutes=true] set to false to hide minutes.
   * @param {Boolean} [cx.showHours=true] set to false to hide hours.
   * @param {Boolean} [cx.showDays=true] set to false to hide days.
   *
   * @example
   *   // IMPORTANT NOTE
   *   //
   *   // If a unit is greater than zero but the show flag for it is set to
   *   // false, then its value will get carried over to the leftmost shown
   *   // unit.
   *
   *   var relTime = baja.RelTime.make({
   *     hours: 1, seconds: 1
   *   });
   *
   *   // 1hour 1sec
   *   relTime.toString();
   *
   *   // 60mins 1sec
   *   relTime.toString({
   *     showHours: false
   *   });
   *
   *   // 3601secs
   *   relTime.toString({
   *     showHours: false,
   *     showMinutes: false
   *   });
   *
   * @returns {String}
   */
  RelTime.prototype.toString = function (cx) {
    cx = cx || {};
    var showMillis = cx.showMillis !== false,
        showSeconds = cx.showSeconds !== false,
        showMinutes = cx.showMinutes !== false,
        showHours = cx.showHours !== false,
        showDays = cx.showDays !== false;

    var relTime = "";
    var ms = this.$ms;

    if (ms < 0) {
      relTime += "-";
      ms = -ms;
    }

    if (ms < 1000) {
      if (showSeconds) {
        if (showMillis) {
          return relTime += ms + "ms";
        } else {
          return "0 sec";
        }
      } else if (showMinutes) {
        return "0 minutes";
      } else if (showHours) {
        return "0 hours";
      } else if (showDays) {
        return "0 days";
      }
    }

    if (ms >= MILLIS_IN_DAY && showDays) {
      var days = Math.abs(this.getDaysPart());
      ms = ms % MILLIS_IN_DAY;
      relTime += days + (days === 1 ? "day " : "days ");
    }

    if (ms >= MILLIS_IN_HOUR && showHours) {
      var hours;

      if (!showDays) {
        hours = Math.abs(this.getHours());
      } else {
        hours = Math.abs(this.getHoursPart());
      }

      ms = ms % MILLIS_IN_HOUR;
      relTime += hours + (hours === 1 ? "hour " : "hours ");
    }

    if (ms >= MILLIS_IN_MINUTE && showMinutes) {
      var mins;

      if (!showHours) {
        mins = Math.abs(this.getMinutes());
      } else {
        mins = Math.abs(this.getMinutesPart());
      }

      ms = ms % MILLIS_IN_MINUTE;
      relTime += mins + (mins === 1 ? "min " : "mins ");
    }

    if (ms > 0 && showSeconds) {
      var secs;

      if (!showMinutes) {
        secs = Math.abs(this.getSeconds());
      } else {
        secs = Math.abs(this.getSecondsPart());
      }

      ms = ms % MILLIS_IN_SECOND;
      relTime += secs;

      var plural = (secs !== 1);

      if (showMillis && ms > 0) {
        relTime += ".";

        if (ms < 10) {
          relTime += "00";
        } else if (ms < 100) {
          relTime += "0";
        }

        relTime += ms;

        plural = plural || (ms !== 0);
      }

      relTime += (plural ? "secs" : "sec");
    }

    return relTime.trim();
  };
  
  /**
   * Return number of milliseconds.
   *
   * @returns {Number}
   */
  RelTime.prototype.getMillis = function () {
    return this.$ms;
  };
  
  /**
   * Return the milliseconds part of this duration.
   * 
   * @returns {Number}
   */
  RelTime.prototype.getMillisPart = function () {
    return this.$ms % 1000;
  };
  
  function truncateToInteger(num) {
    return Math[num < 0 ? 'ceil' : 'floor'](num);
  }
  /**
   * Return number of seconds.
   *
   * @returns {Number}
   */
  RelTime.prototype.getSeconds = function () {
    return truncateToInteger(this.$ms / MILLIS_IN_SECOND);
  };
  
  /**
   * Return the seconds part of this duration.
   * 
   * @returns {Number}
   */
  RelTime.prototype.getSecondsPart = function () {
    return this.getSeconds() % 60;
  };
  
  /**
   * Return number of minutes.
   *
   * @returns {Number}
   */
  RelTime.prototype.getMinutes = function () {
    return truncateToInteger(this.$ms / MILLIS_IN_MINUTE);
  };
  
  /**
   * Return the minutes part of this duration.
   * 
   * @returns {Number}
   */
  RelTime.prototype.getMinutesPart = function () {
    return this.getMinutes() % 60;
  };
  
  /**
   * Return number of hours.
   *
   * @returns {Number}
   */
  RelTime.prototype.getHours = function () {
    return truncateToInteger(this.$ms / MILLIS_IN_HOUR);
  };
  
  /**
   * Return the hours part of this duration.
   * 
   * @returns {Number}
   */
  RelTime.prototype.getHoursPart = function () {
    return this.getHours() % 24;
  };
  
  /**
   * Return number of days.
   *
   * @returns {Number}
   */
  RelTime.prototype.getDays = function () {
    return truncateToInteger(this.$ms / MILLIS_IN_DAY);
  };
  
  /**
   * Return the days part of this duration.
   * 
   * @returns {Number}
   */
  RelTime.prototype.getDaysPart = function () {
    return this.getDays();
  };
  
  
  return RelTime;
});