/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/**
 * Defines {@link baja.OrdQueryList}.
 * @module baja/ord/OrdQueryList
 */
define(["bajaScript/sys",
        "bajaScript/baja/ord/OrdQueryListCursor"], 
    function (baja, OrdQueryListCursor) {
  
  "use strict";
  
  var subclass = baja.subclass,
      bajaDef = baja.def,
      BaseBajaObj = baja.BaseBajaObj;
  
  /**
   * ORD Query List.
   * 
   * Used to hold a list of OrdQueries.
   *
   * @class
   * @alias baja.OrdQueryList
   * @extends baja.BaseBajaObj
   *
   * @see baja.OrdQuery
   *
   * @param {Array.<baja.OrdQuery>} [queries] an array of ORD queries.
   */  
  var OrdQueryList = function OrdQueryList(queries) {  
    this.$queries = queries || [];
  };
  
  subclass(OrdQueryList, BaseBajaObj);
  
  /**
   * Add an ORD Query to the List
   *
   * @param {baja.OrdQuery} query 
   */
  OrdQueryList.prototype.add = function (query) {
    this.$queries.push(query); 
    this.$hasUnknown = undefined; // reset the cached hasUnknown result
  };
  
  /**
   * Is the list empty?
   *
   * @returns {Boolean}
   */
  OrdQueryList.prototype.isEmpty = function () {
    return this.$queries.length === 0;
  };
  
  /**
   * Does the list contain an unknown ORD scheme?
   *
   * @private
   *
   * @returns {Boolean}
   */
  OrdQueryList.prototype.hasUnknown = function () {
    if (this.$hasUnknown !== undefined) {
      return this.$hasUnknown;
    }
  
    // Search for an unknown ORD scheme and cache the result
    var i,
        unknown = false;
        
    for (i = 0; i < this.$queries.length; ++i) {
      if (this.$queries[i].getScheme() instanceof baja.UnknownScheme) {
        unknown = true;
      }
    }
    this.$hasUnknown = unknown;
    return unknown;
  };
  
  /**
   * Returns a Cursor for use with the ORD Query List.
   *
   * @returns {module:baja/ord/OrdQueryListCursor}
   */
  OrdQueryList.prototype.getCursor = function () {
    return new OrdQueryListCursor(this);
  };
  
  function normalizeOrdQueryList(list) {
    var i;
    for (i = 0; i < list.$queries.length; ++i) {
      if (list.$queries[i].normalize(list, i)) {
        return true;
      }
    }
    return false;
  }
  
  /**
   * Normalize the ORD Query List.
   * 
   * @private
   *
   * @returns {baja.OrdQueryList} returns itself.
   */
  OrdQueryList.prototype.normalize = function () {
    // Don't try to normalize if there's unknown client ORD Schemes
    if (!this.hasUnknown()) {
      for (;;) {
        if (!normalizeOrdQueryList(this)) {
          break;
        }
      }
    }
    return this;
  };
  
  /**
   * Return String representation of the ORD Query List.
   *
   * @param {Number} [length=size()] number of queries to include in the string 
   * @returns {String}
   */
  OrdQueryList.prototype.toString = function (length) {
    length = bajaDef(length, this.$queries.length);
    var a = [], i;
    for (i = 0; i < length; ++i) {
      a.push(this.$queries[i].toString());
    }
    return a.join("|");
  };
  
  /**
   * Return the query object at the specified index.
   *
   * @param {Number|String} index or scheme name.
   * @returns {baja.OrdQuery} query (or null if can't be found).
   */
  OrdQueryList.prototype.get = function (index) {
    var to = typeof index,
        queries = this.$queries,
        i,
        q;
        
    if (to === "number") {
      // Get via index
      q = queries[index];
    }
    else if (to === "string") {
      // Search via scheme name
      for (i = 0; i < queries.length; ++i) {
        if (queries[i].getSchemeName() === index) {
          q = queries[i];
          break;
        }
      }
    }
    
    return q || null;
  };
    
  /**
   * Set an ORD query object at the given index.
   *
   * @param {Number} index 
   * @param {baja.OrdQuery} query
   */
  OrdQueryList.prototype.set = function (index, query) {
    if (index < 0 || index > this.$queries.length) {
      throw new Error("Invalid index (" + index + ")");
    }
    this.$queries[index] = query;
    this.$hasUnknown = undefined; // reset the cached hasUnknown result
  };
  
  /**
   * Remove the entry at the specified index and return it.
   *
   * @param {Number} index 
   * @returns {baja.OrdQuery} query
   */
  OrdQueryList.prototype.remove = function (index) {
    var query = null;
    if (index >= 0 && index < this.$queries.length) {
      query = this.$queries.splice(index, 1)[0];
      this.$hasUnknown = undefined; // reset the cached hasUnknown result    
    }
    return query;
  };
  
  /**
   * Return the size of the query list.
   * 
   * @returns {Number} size of the list.
   */
  OrdQueryList.prototype.size = function () {
    return this.$queries.length;
  };
  
  return OrdQueryList;
});