/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Danesh Kamal
 */

/**
 * @module nmodule/alarm/rc/db/editors/AlarmTable
 * @private
 */
define(['baja!',
  'baja!alarm:AlarmRecord,bql:DynamicTimeRange,bql:AbsTimeFilter',
  'jquery',
  'underscore',
  'Promise',
  'nmodule/webEditors/rc/fe/fe',
  'nmodule/alarm/rc/Alarm',
  'nmodule/alarm/rc/util/alarmUtils',
  'nmodule/webEditors/rc/wb/table/Table',
  'nmodule/webEditors/rc/wb/table/model/Column',
  'nmodule/webEditors/rc/wb/table/model/columns/PropertyColumn',
  'nmodule/alarm/rc/console/table/columns/InfoColumn',
  'nmodule/alarm/rc/db/AlarmPropertyColumn',
  'nmodule/alarm/rc/db/AlarmFacetColumn',
  'nmodule/alarm/rc/db/commands/AlarmDetailsCommand',
  'nmodule/webEditors/rc/wb/table/model/Row',
  'nmodule/webEditors/rc/wb/table/model/ComponentTableModel',
  'nmodule/webEditors/rc/wb/table/model/source/ArrayComponentSource',
  'nmodule/webEditors/rc/wb/table/menu/ShowHideColumnsContextMenu'
], function (baja,
             types,
             $,
             _,
             Promise,
             fe,
             Alarm,
             alarmUtils,
             Table,
             Column,
             PropertyColumn,
             InfoColumn,
             AlarmPropertyColumn,
             AlarmFacetColumn,
             AlarmDetailsCommand,
             Row,
             ComponentTableModel,
             ArrayComponentSource,
             ShowHideColumnsContextMenu) {

  "use strict";

  var AlarmTable = function () {
  };

  AlarmTable.makeFor = function (view, alarms) {

    var ALARM_TABLE_SELECTOR       = '.alarmTable',
        ALARM_COUNT_LABEL_SELECTOR = '.alarmCount',
        TABLE_CONTAINER_SELECTOR   = '.tableContainer',
        COLUMN_FILTER_SELECTOR     = '.columnFilter',
        COLUMN_FLAGS_CHANGED_EVENT = 'columnFlagsChanged',
        alarmTable                 = $(ALARM_TABLE_SELECTOR, view.jq()).data('widget'),
        alarmDetailsCmd,
        visibleCols                = [
          'timestamp',
          'sourceState',
          'ackState',
          'source',
          'alarmClass',
          'priority',
          'msgText'
        ];

    function makeColumn(obj, map) {

      //obj is either a Property or Facet key
      if (!(obj instanceof baja.Property || typeof obj === 'string')) {
        return new Column("#Error");
      }

      var col, prevCol,
          model,
          isProperty = obj instanceof baja.Property;

      col = isProperty ? new AlarmPropertyColumn(obj, map, {displayName: obj.$getDisplayName()}) :
        new AlarmFacetColumn(obj, {displayName: alarmUtils.getAlarmDataDisplayName(obj)});

      if (alarmTable) {
        model = alarmTable.value();
        prevCol = model.getColumn(col.getName());
        if (prevCol) {
          col.setUnseen(prevCol.isUnseen());
        }
      }

      else if ($.inArray(col.getName(), visibleCols) === -1) {
        col.setUnseen(true);
      }

      return col;
    }

    function sortTableByColumn(table, column, descending) {

      function getSortColumnObj() {
        var model = table.value(),
            th    = table.jq().find('thead > th').filter(function () {
              var $this = $(this);
              return $this.hasClass('asc') || $this.hasClass('desc');
            }).first(),
            col   = th.data('column');
        col = col || model.getColumn('timestamp');
        return {col: col, desc: th.data('desc')};
      }

      if (!column) {
        var obj = getSortColumnObj();
        column = obj.col;
        descending = obj.desc;
      }

      return Promise.resolve(view.pager && view.pager.update({
          desc: descending,
          col: column
        })).then(function (result) {
        return !result && table.$sortByColumnDisplay(column, descending);
      }).then(function (result) {
        return result ? table : view.table;
      }).then(function (table) {
        var headers = table.$getThead().children('th');
        headers.removeClass('asc desc').data('desc', false);
        var th = headers.filter(function () {
          return $(this).data('column').getName() === column.getName();
        }).first();
        th.addClass(descending ? 'desc' : 'asc').data('desc', descending);
        return table;
      });
    }

    function armEventHandlers(table) {
      armAlarmDetails(table);
      armColumnSorting(table);
      armColumnFilter(table);
      armColumnSizing(table);
      return table;
    }

    function armAlarmDetails(table) {
      //arm double click functionality to launch alarm details dialog
      $("tr", table.jq()).dblclick(function () {

        if (alarmDetailsCmd) {
          alarmDetailsCmd.invoke();
        }

        else {
          var filteredGroup = view.getCommandGroup().filter({
            include: function (cmd) {
              return cmd instanceof AlarmDetailsCommand;
            }
          });

          alarmDetailsCmd = !filteredGroup.isEmpty() && filteredGroup.get(0);

          if (alarmDetailsCmd) {
            alarmDetailsCmd.invoke();
          }
        }
      });
    }

    function armColumnSorting(table) {

      if (!table.sortByColumn) {
        table.sortByColumn = function (column, descending) {
          return sortTableByColumn(this, column, descending);
        };
      }

      //arm table column sorting
      table.jq().off('click', 'thead > th').on('click', 'thead > th', function () {
        var $this = $(this); //particular th that was clicked
        if ($this.data('column').isSortable()) {
          table.sortByColumn($this.data('column'), !$this.data('desc'));
        }
      });
    }

    function armColumnFilter(table) {
      //add and arm Column filter
      var dom = view.jq();
      $(COLUMN_FILTER_SELECTOR, dom).off(); //turn off existing event handler
      new ShowHideColumnsContextMenu(table).arm(dom, COLUMN_FILTER_SELECTOR);
    }

    function armColumnSizing(table) {
      //set width of last visible column to 100% when column filter is modified
      var model = table.value();
      model.on(COLUMN_FLAGS_CHANGED_EVENT, function () {
        sizeVisibleColumns(table);
      });
    }

    function sizeVisibleColumns(table) {
      function getColumnDom(column) {
        return table.jq().find('.js-col-' + column.getName());
      }

      var model          = table.value(),
          visibleColumns = _.filter(model.getColumns(), function (col) {
            return !col.isUnseen();
          }),
          length         = visibleColumns.length;

      _.each(visibleColumns, function (col, index) {
        getColumnDom(col).css('width', index === length - 1 ? '100%' : 'auto');
      });

      return table;
    }

    function loadExistingTable(model) {
      return Promise.resolve(alarmTable && alarmTable.load(model));
    }

    function loadNewTable(model) {
      var dom = $('<table class="alarmTable ux-table noselect"></table>');
      return fe.buildFor({
        dom: dom,
        value: model,
        type: Table
      }).then(function (table) {
        dom.appendTo($(TABLE_CONTAINER_SELECTOR, view.jq()));
        alarmTable = table;
      });
    }

    function loadTable(model) {
      return loadExistingTable(model).then(function (loaded) {
        return Promise.resolve(!loaded && loadNewTable(model));
      }).then(function () {
        view.table = alarmTable;
        return alarmTable;
      });
    }

    function makeModel(map) {
      var columns     = [],
          alarmCount  = alarms.length,
          alarmRecord = baja.$('alarm:AlarmRecord'),
          facetKeys   = ['msgText'];

      //add info/icons column
      columns.push(new InfoColumn());

      //add standard alarm property cols
      alarmRecord.getSlots().properties().each(function (property) {
        columns.push(makeColumn(property, map));
      });

      //add facet columns
      facetKeys.forEach(function (key) {
        columns.push(makeColumn(key, map));
      });

      $(ALARM_COUNT_LABEL_SELECTOR, view.jq())
        .html(alarmCount + (alarmCount > 1 || alarmCount === 0 ? ' Alarms' : ' Alarm'));

      var model = new ComponentTableModel({
        componentSource: new ArrayComponentSource({container: alarms}),
        columns: columns
      });

      return Promise.resolve(model);
    }

    return Alarm.getDisplayNamesMap()
      .then(makeModel)
      .then(loadTable)
      .then(armEventHandlers)
      .then(sizeVisibleColumns);
  };

  return AlarmTable;
});
