/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Danesh Kamal
 */

/**
 * @module nmodule/alarm/rc/db/editors/Pager
 * @private
 */
define([
    'baja!',
    'baja!bql:DynamicTimeRange',
    'lex!alarm',
    'jquery',
    'Promise',
    'dialogs',
    'nmodule/webEditors/rc/fe/fe',
    'nmodule/alarm/rc/Alarm',
    'bajaux/commands/Command',
    'bajaux/util/CommandButton',
    'nmodule/alarm/rc/db/editors/AlarmTable'
  ],
  function (baja, types, lexs, $, Promise, dialogs, fe, Alarm, Command, CommandButton, AlarmTable) {

    "use strict";

    var lex                        = lexs[0],
        TIME_RANGE_EDITOR_SELECTOR = '.timeRangeEditor',
        ALARM_LIMIT_SELECTOR       = '.limitSelector',
        PREV_PAGE_CMD_SELECTOR     = '.prevPageCmd',
        NEXT_PAGE_CMD_SELECTOR     = '.nextPageCmd',
        PAGE_LOCATION_SELECTOR     = '.pageLocation',
        MAINTENANCE_GROUP_SELECTOR = '.maintenance',
        PAGING_SELECTOR            = '.paging',
        NEXT_PAGE_LABEL            = lex.get('command.nextPage'),
        PREVIOUS_PAGE_LABEL        = lex.get('command.previousPage'),
        LIMITS                     = [10, 25, 50, 75, 100, 250, 500, 1000],
        DEFAULT_LIMIT              = LIMITS[4];

    function initLimitSelector(pager) {
      var view       = pager.view,
          selectList = $(ALARM_LIMIT_SELECTOR, view.jq());

      selectList.empty();
      LIMITS.forEach(function (limit) {
        var option = $("<option>").val(limit).html(limit + " per page");
        selectList.append(option);
      });

      selectList.val(DEFAULT_LIMIT);

      selectList.change(function () {
        pager.offset = 0;
        pager.limit = parseInt(selectList.val());
        pager.updateTable();
      });
    }

    function init(pager) {

      initLimitSelector(pager);

      return Promise.join(makePrevCommand(pager), makeNextCommand(pager), Alarm.getDisplayNamesMap(), baja.TimeZoneDatabase.get(),
        function (prev, next, alarmClassDisplayNames, tzDatabase) {
          pager.prev = prev;
          pager.next = next;
          pager.alarmClassDisplayNames = alarmClassDisplayNames;
          pager.timezoneDatabase = tzDatabase;
          return pager.update();
        });
    }

    function makePrevCommand(pager) {
      return fe.buildFor({
        dom: $(PREV_PAGE_CMD_SELECTOR, pager.view.jq()),
        type: CommandButton,
        value: new Command({
          displayName: PREVIOUS_PAGE_LABEL,
          func: function () {
            return pager.previousPage();
          }
        })
      });
    }

    function makeNextCommand(pager) {
      return fe.buildFor({
        dom: $(NEXT_PAGE_CMD_SELECTOR, pager.view.jq()),
        type: CommandButton,
        value: new Command({
          displayName: NEXT_PAGE_LABEL,
          func: function () {
            return pager.nextPage();
          }
        })
      });
    }

    /**
     * Disable/Enable controls when table starts/stops loading
     *
     * @param pager
     * @param isLoading
     */
    function loading(pager, isLoading) {
      var view                  = pager.view,
          timeRangeControl      = $(TIME_RANGE_EDITOR_SELECTOR + " select", view.jq()),
          resultsPerPageControl = $(PAGING_SELECTOR + " select", view.jq()),
          maintenanceControls   = $(MAINTENANCE_GROUP_SELECTOR + " button", view.jq()),
          controls              = maintenanceControls.add(timeRangeControl).add(resultsPerPageControl);

      controls.prop('disabled', isLoading);
    }

    var Pager = function (view) {
      this.view = view;
      this.limit = DEFAULT_LIMIT;
      this.offset = 0;
      this.filterSet = null;
    };

    Pager.prototype.$getLimitSelector = function () {
      return $(ALARM_LIMIT_SELECTOR, this.view.jq());
    };

    Pager.prototype.$getPreviousPageCmd = function () {
      return this.prev;
    };

    Pager.prototype.$getNextPageCmd = function () {
      return this.next;
    };

    Pager.prototype.updateTable = function () {
      var table = this.view.table;
      return table && table.sortByColumn && table.sortByColumn();
    };

    Pager.prototype.previousPage = function () {
      this.offset -= this.limit;
      this.offset = this.offset < 0 ? 0 : this.offset;
      return this.updateTable();
    };

    Pager.prototype.nextPage = function () {
      this.offset += this.limit;
      return this.updateTable();
    };

    Pager.prototype.update = function (args) {
      var that            = this,
          view            = that.view,
          timeRangeEditor = view.timeRangeEditor,
          limit           = that.limit,
          offset          = that.offset,
          filterSet       = that.filterSet,
          desc            = args && args.desc,
          col             = args && args.col && args.col.getName() || 'timestamp',
          DELAY           = 1500;

      if (!timeRangeEditor) {
        return Promise.resolve(that);
      }

      return this.disable().then(function () {
        return timeRangeEditor.read();
      }).then(function (timeRange) {
        //query the alarm database
        loading(that, true);
        var promise = Alarm.retrieveAlarms(timeRange, limit, offset, filterSet, desc, col, that.alarmClassDisplayNames, that.timezoneDatabase);
        dialogs.showLoading(DELAY, promise);
        return promise;
      }).then(function (obj) {
        //build the alarm table and return the page count
        var alarms     = obj && obj.alarms || [],
            alarmCount = obj && obj.count || 0,
            promise    = AlarmTable.makeFor(view, alarms).then(function () {
              return Math.ceil(alarmCount / that.limit);
            });
        dialogs.showLoading(DELAY, promise);
        return promise;
      }).then(function (pageCount) {
        //update the next/prev button states and page position based on the page count
        var currentPage = that.getCurrentPage();
        that.next.setEnabled(currentPage < pageCount);
        that.prev.setEnabled(currentPage > 1 && pageCount >= currentPage);
        that.displayPagePosition(currentPage, pageCount);
        return null;
      }).then(function () {
        loading(that, false);
        return that;
      });
    };

    Pager.prototype.getAlarmCount = function (timeRange) {
      timeRange = timeRange || baja.$('bql:DynamicTimeRange').make("timeRange");
      return Alarm.getAlarmCount(timeRange, this.filterSet);
    };

    Pager.prototype.getPageCount = function (timeRange) {
      var that = this;
      return that.getAlarmCount(timeRange).then(function (count) {
        return Math.ceil(count / that.limit);
      });
    };

    Pager.prototype.getCurrentPage = function () {
      return 1 + this.offset / this.limit;
    };

    Pager.prototype.displayPagePosition = function (currentPage, totalPages) {

      var jq = $(PAGE_LOCATION_SELECTOR, this.view.jq());

      if (totalPages === 0) {
        jq.html("No results");
      }

      else if (currentPage <= totalPages) {
        jq.html("Page " + currentPage + " of " + totalPages);
      }
    };

    Pager.prototype.disable = function () {
      return Promise.resolve(this.prev && this.prev.setEnabled(false) &&
        this.next && this.next.setEnabled(false));
    };

    Pager.prototype.destroy = function () {
      return Promise.join(this.prev.destroy(), this.next.destroy(),
        this.alarmClassDisplayNames.destroy(), this.timezoneDatabase.destroy());
    };

    Pager.makeFor = function (view) {
      var pager = new Pager(view);
      return init(pager);
    };

    return Pager;
  });
