/**
 * @copyright 2017 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/**
 * Manages communications for the browser environment.
 *
 * @module env/BrowserCommsManager
 * @private
 */
define(["bajaBrowserEnvUtil", 
        "bajaScript/comm", 
        "bajaScript/env/ConnectionManager"], function (
        envUtil, 
        baja, 
        ConnectionManager) {

  "use strict";

  /**
   * Manages communications for the browser environment.
   *
   * @class
   * @private
   * */
  var BrowserCommsManager = function () {
    var that = this,
        asyncQueue = that.$asyncQueue = [];

    /**
     * Asynchronously processes the next message in the queue.
     *
     * @private
     */
    that.processNext = function () {
      // Remove the first element the queue
      asyncQueue.shift();
    
      if (asyncQueue.length > 0) {
        // Remove and return the first element of the async queue
        that.sendAsync(asyncQueue[0]);
      }
    };

    that.$connectionManager = new ConnectionManager(that.processNext);
  };

  /**
   * Sends the message asynchronously.
   * 
   * @param {Object} frameData The frame data to send.
   * @private
   */
  BrowserCommsManager.prototype.sendAsync = function (frameData) {
    // Flag that we're sending the data
    frameData.$sending = true;
  
    // If we can't send the frame through a connection then send it via HTTP
    if (!this.$connectionManager.send(frameData)) {
      // Make the HTTP POST to the BOX Servlet
      envUtil.sendHttp("POST", 
                       "/box/", 
                       /*async*/true, 
                       frameData.cb, 
                       frameData.frame.toString(),
                       frameData.processed /*may be undefined*/);
    }
  };

  /**
   * In a browser, <code>baja.comm.BoxFrame#send</code> will delegate this method.
   *
   * @private
   * @see module:baja/comm/BoxFrame
   *
   * @param {Object} boxFrame The box frame to send.
   * @param {baja.comm.Batch|baja.comm.Callback|Object} callback an object
   * containing ok/fail callbacks.
   */
   BrowserCommsManager.prototype.sendBoxFrame = function (boxFrame, callback) {
    var that = this,
        frameData,
        processNext = that.processNext,
        asyncQueue = that.$asyncQueue;
  
    if (!boxFrame.$sync) {
      frameData = {
        frame: boxFrame,
        cb: callback
      };

      if (boxFrame.$queue) {
        frameData.processed = processNext;

        // Push the frame onto our async queue (adds to end of array).
        asyncQueue.push(frameData);
        
        // If there are no outstanding frames then simply process the queue (otherwise
        // rely on comms to make processNext request).
        if (asyncQueue.length === 1) {
          that.sendAsync(asyncQueue[0]); 
        }
      }
      else {
         that.sendAsync(frameData);
      }
    }
    else {
      // For synchronous network calls, always make an HTTP call.
      envUtil.sendHttp("POST", 
                       "/box/", 
                       /*async*/false, 
                       callback, 
                       boxFrame.toString(),
                       processNext);
    }
  };

  /**
   * Returns the connection manager.
   *
   * @private
   * 
   * @returns {module:env/ConnectionManager} The connection manager.
   */
  BrowserCommsManager.prototype.getConnectionManager = function () {
    return this.$connectionManager;
  };
                          
  return BrowserCommsManager;
});  