/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/AbsTimeEditor
 */
define(['baja!',
        'jquery',
        'Promise',
        'bajaux/events',
        'nmodule/webEditors/rc/fe/baja/BaseEditor',
        'nmodule/webEditors/rc/fe/baja/DateEditor',
        'nmodule/webEditors/rc/fe/baja/TimeEditor'], function (
         baja,
         $,
         Promise,
         events,
         BaseEditor,
         DateEditor,
         TimeEditor) {

  'use strict';

  var MODIFY_EVENT = events.MODIFY_EVENT;

  /**
   * A field editor for working with AbsTimes.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/baja/BaseEditor
   * @alias module:nmodule/webEditors/rc/fe/baja/AbsTimeEditor
   */
  var AbsTimeEditor = function AbsTimeEditor(params) {
    BaseEditor.apply(this, arguments);
    this.$dateEditor = new DateEditor(params);
    this.$timeEditor = new TimeEditor(params);
  };
  AbsTimeEditor.prototype = Object.create(BaseEditor.prototype);
  AbsTimeEditor.prototype.constructor = AbsTimeEditor;

  /**
   * Initializes `DateEditor` and `TimeEditor` for the date and time parts of
   * the `AbsTime`. Depending on the state of the `showDate` and `showTime`
   * facets, these editors may be hidden in the DOM.
   *
   * @param {JQuery} dom
   * @returns {Promise} promise to be resolved when both date and time
   * editors have finished initializing
   */
  AbsTimeEditor.prototype.doInitialize = function (dom) {
    //TODO: time zones :(

    var that = this,
        props = that.properties(),
        showDate = props.getValue('showDate') !== false,
        showTime = props.getValue('showTime') !== false,
        dateDiv = $('<div class="date"/>').appendTo(dom),
        timeDiv = $('<div class="time"/>').appendTo(dom);

    if (!showDate) {
      dateDiv.hide();
    }

    if (!showTime) {
      timeDiv.hide();
    }

    if (!showDate && !showTime) {
      throw new Error('both showDate and showTime facets set to false');
    }

    dom.on(MODIFY_EVENT, '.editor', function () {
      that.setModified(true);
      return false;
    });

    return Promise.join(that.$dateEditor.initialize(dateDiv),
                        that.$timeEditor.initialize(timeDiv));
  };

  /**
   * Loads the date and time parts of the `AbsTime` into the date and time
   * editors.
   *
   * @param {baja.AbsTime} value
   * @returns {Promise} promise to be resolved once the date and time
   * editors have finished loading
   */
  AbsTimeEditor.prototype.doLoad = function (value) {
    return Promise.join(this.$dateEditor.load(value.getDate()),
                        this.$timeEditor.load(value.getTime()));
  };

  /**
   * Reads the values from the date and time editors and uses them to construct
   * a new `AbsTime` instance.
   *
   * @returns {Promise} promise to be resolved with a `baja.AbsTime`
   * constructed from the currently entered date and time, or rejected if the
   * date/time editors cannot be read
   */
  AbsTimeEditor.prototype.doRead = function () {
    //TODO: AbsTime offset
    return Promise.join(this.$dateEditor.read(), this.$timeEditor.read())
      .spread(function (date, time) {
        return baja.AbsTime.make({
          date: date,
          time: time
        });
      });
  };

  /**
   * Enables/disables both the date and time editors.
   *
   * @param {Boolean} enabled
   */
  AbsTimeEditor.prototype.doEnabled = function (enabled) {
    return this.getChildEditors().setAllEnabled(enabled);
  };

  /**
   * Readonly/writable both the date and time editors.
   *
   * @param {Boolean} readonly
   */
  AbsTimeEditor.prototype.doReadonly = function (readonly) {
    return this.getChildEditors().setAllReadonly(readonly);
  };

  /**
   * Destroys the date and time editors.
   */
  AbsTimeEditor.prototype.doDestroy = function () {
    return this.getChildEditors().destroyAll();
  };

  return (AbsTimeEditor);
});
