/*
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 * @author Vikram Nagulan
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/ConfirmPasswordEditor
 */
define(['baja!',
        'lex!webEditors',
        'jquery',
        'nmodule/webEditors/rc/fe/baja/PasswordEditor',
        'nmodule/webEditors/rc/fe/CompositeEditor',
        'nmodule/webEditors/rc/fe/config/CompositeBuilder'], function (
         baja,
         lexs,
         $,
         PasswordEditor,
         CompositeEditor,
         CompositeBuilder) {

  'use strict';

  var webEditorsLex = lexs[0],
    CONFIRM = webEditorsLex.get('UserPasswordEditor.confirm'),
    PASSWORD = webEditorsLex.get('UserPasswordEditor.password'),
    PASSWORD_MISMATCH = webEditorsLex.get('UserPasswordEditor.passwordMismatch'),
    KEY_PASSWORD = 'p',
    KEY_CONFIRM = 'c',
    keys = [KEY_PASSWORD, KEY_CONFIRM];

  /**
   * Editor for handling `baja:Password` values that need a confirmation.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/CompositeEditor
   * @alias module:nmodule/webEditors/rc/fe/baja/ConfirmPasswordEditor
   */
  var ConfirmPasswordEditor = function ConfirmPasswordEditor(params) {
    var that = this;
    CompositeEditor.apply(this, arguments);

    that.validators().add(function (val) {
      return that.$getConfirmEditor().read()
        .then(function (confirm) {
          if (val.encodeToString() !== confirm.encodeToString()) {
            throw new Error(PASSWORD_MISMATCH);
          }
        });
    });
  };
  ConfirmPasswordEditor.prototype = Object.create(CompositeEditor.prototype);
  ConfirmPasswordEditor.prototype.constructor = ConfirmPasswordEditor;

  ConfirmPasswordEditor.prototype.doInitialize = function (dom) {
    return CompositeEditor.prototype.doInitialize.apply(this, arguments);
  };
  /**
   * @overrides module:nmodule/webEditors/rc/fe/CompositeEditor
   * @returns {*}
   */
  ConfirmPasswordEditor.prototype.makeBuilder = function () {
    var that = this;
    var builder = new CompositeBuilder();

    builder.getDomFor = function (i) {
      return $('<div data-key="' + i + '"></div>').appendTo(that.jq());
    };
    builder.getKeys = function () {
      return keys;
    };

    builder.getValueFor = function () {
      return baja.$("baja:Password");
    };

    builder.getConfigFor = function (i) {
      var label = "";
      switch (i) {
        case KEY_PASSWORD:
          label = PASSWORD;
          break;
        case KEY_CONFIRM:
          label = CONFIRM;
          break;
        default:
          label = "";
      }
      return {
        type: PasswordEditor,
        complex: that.getComplex(),
        slot: that.getSlot(),
        properties: {prelabel: label}
      };
    };

    return builder;
  };

  /**
   * Get the second/confirmation string editor.
   *
   * @private
   * @returns {module:nmodule/webEditors/rc/fe/baja/PasswordEditor}
   */
  ConfirmPasswordEditor.prototype.$getConfirmEditor = function () {
    // noinspection JSValidateTypes
    return this.getBuilder().getEditorFor(KEY_CONFIRM);
  };

  ConfirmPasswordEditor.prototype.doRead = function () {
    return this.getBuilder().readAll().spread(function (pwd) {
      return pwd;
    });
  };

  /**
   * Save the password
   *
   * @see module:nmodule/webEditors/rc/fe/baja/PasswordEditor
   * @see module:nmodule/webEditors/rc/servlets/password
   * @returns {Promise}
   */
  ConfirmPasswordEditor.prototype.saveToComplex = function (pw, params) {
    return PasswordEditor.prototype.saveToComplex.call(this, pw, params);
  };

  /**
   * Destroys "password" and "confirm" editors.
   *
   * @returns {Promise}
   */
  ConfirmPasswordEditor.prototype.doDestroy = function () {
    return this.getChildWidgets().destroyAll();
  };

  /**
   * Set password/confirm editors readonly. Will not make them writable if
   * editor is not working in a secure environment.
   *
   * @see module:nmodule/webEditors/rc/fe/baja/PasswordEditor
   * @param {Boolean} readonly
   * @returns {Promise}
   */
  ConfirmPasswordEditor.prototype.doReadonly = function (readonly) {
    return this.getChildEditors().setAllReadonly(readonly);
  };

  /**
   * Set password/confirm editors enabled. Will not enable them if editor is
   * not working in a secure environment.
   *
   * @see module:nmodule/webEditors/rc/fe/baja/PasswordEditor
   * @param {Boolean} enabled
   * @returns {Promise}
   */
  ConfirmPasswordEditor.prototype.doEnabled = function (enabled) {
    return this.getChildEditors().setAllEnabled(enabled);
  };

  return (ConfirmPasswordEditor);
});
