/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/DisplayOnlyEditor
 */
define(['baja!',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/baja/BaseEditor',
        'nmodule/webEditors/rc/util/htmlUtils'], function (
         baja,
         Promise,
         _,
         BaseEditor,
         htmlUtils) {

  'use strict';

  var escapeHtml = htmlUtils.escapeHtml;

  /**
   * A display-only field editor. This editor will do nothing but convert the
   * value to a string and set the text value of the bound element.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/baja/BaseEditor
   * @alias module:nmodule/webEditors/rc/fe/baja/DisplayOnlyEditor
   */
  var DisplayOnlyEditor = function DisplayOnlyEditor() {
    BaseEditor.apply(this, arguments);
  };
  DisplayOnlyEditor.prototype = Object.create(BaseEditor.prototype);
  DisplayOnlyEditor.prototype.constructor = DisplayOnlyEditor;

  /**
   * Make an editor display-only. `doInitialize()`, `doLoad()`, and `doSave()`
   * will always be overridden with display-only methods. If `updateDisplay()`
   * or `valueToString()` are implemented by the editor, they will be
   * preserved.
   *
   * @private
   * @param {module:nmodule/webEditors/rc/fe/baja/BaseEditor} ed
   */
  DisplayOnlyEditor.$mixin = function (ed) {
    var prot = DisplayOnlyEditor.prototype;
    ed.doInitialize = prot.doInitialize;
    ed.doLoad = prot.doLoad;
    ed.doRead = prot.doRead;
    ed.doSave = prot.doSave;
    ed.doDestroy = prot.doDestroy;
    ed.updateDisplay = ed.updateDisplay || prot.updateDisplay;
    ed.valueToString = ed.valueToString || prot.valueToString;
    ed.$mixins = _.without(ed.$mixins, 'batchSave');
  };

  /**
   * Converts the value to a string using the standard `String()` function.
   * Override to add special to-string logic.
   *
   * @param {*} value
   * @returns {String|Promise<String>}
   */
  DisplayOnlyEditor.prototype.valueToString = function (value) {
    if (baja.hasType(value, 'baja:Simple')) {
      var cx = this.properties().toValueMap();
      return Promise.resolve(value.toString(cx));
    }
    return String(value);
  };

  /**
   * Updates the current display. Useful if a `baja.Complex` has loaded and
   * you wish to update the text display with new changes. Especially useful
   * with `bajaux/mixin/subscriberMixIn` - just call it on each `changed` event.
   *
   * @param {*} value
   * @returns {Promise|undefined}
   */
  DisplayOnlyEditor.prototype.updateDisplay = function (value) {
    var that = this;

    if (typeof value === 'undefined') {
      value = that.value();
    }

    return Promise.resolve(that.valueToString(value))
      .then(function (str) {
        that.jq().html(escapeHtml(str).replace(/\n/g, '<br>'));
      });
  };

  /**
   * Converts the value to a string and sets the element's text value.
   * @param {*} value
   */
  DisplayOnlyEditor.prototype.doLoad = function (value) {
    return this.updateDisplay(value);
  };

  return (DisplayOnlyEditor);
});
