/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/DynamicEnumEditor
 */
define(['baja!',
        'jquery',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/CompositeEditor',
        'nmodule/webEditors/rc/fe/baja/util/rangeUtils',
        'nmodule/webEditors/rc/fe/config/CompositeBuilder'], function (
         baja,
         $,
         Promise,
         _,
         CompositeEditor,
         rangeUtils,
         CompositeBuilder) {

  'use strict';

  var getEnumRangeDisplay = rangeUtils.getEnumRangeDisplay;

  /**
   * A field editor for working with `DynamicEnum`s.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/CompositeEditor
   * @alias module:nmodule/webEditors/rc/fe/baja/DynamicEnumEditor
   */
  var DynamicEnumEditor = function DynamicEnumEditor(params) {
    CompositeEditor.call(this, $.extend({
      keyName: 'DynamicEnumEditor'
    }, params));
  };
  DynamicEnumEditor.prototype = Object.create(CompositeEditor.prototype);
  DynamicEnumEditor.prototype.constructor = DynamicEnumEditor;

  /**
   * Use a single-key CompositeBuilder so I don't have to manage all those
   * events by hand.
   * TODO: a ProxyEditor or similar because this is such a common use case -
   * i have a value of one data type, but i want to translate it into this
   * other type for the editor
   *
   * @private
   * @returns {module:nmodule/webEditors/rc/fe/config/CompositeBuilder}
   */
  DynamicEnumEditor.prototype.makeBuilder = function () {
    var that = this,
        builder = new CompositeBuilder();

    builder.getKeys = function () { return [ 'string' ]; };

    builder.getDomFor = function () { return $('<div/>').appendTo(that.jq()); };

    builder.getValueFor = function () {
      var dynEnum = this.getDataSource();
      return getEnumRangeDisplay(dynEnum.getOrdinal(), dynEnum.getRange());
    };

    builder.getConfigFor = function () {
      var range = that.$getRangeFacet() || this.getDataSource().getRange(),
          ordinals = range.getOrdinals(),
          datalist = {};

      return Promise.all(_.map(ordinals, function (ordinal) {
        return getEnumRangeDisplay(ordinal, range);
      }))
        .then(function (displays) {
          _.each(ordinals, function (ordinal, i) {
            datalist[ordinal] = displays[i];
          });
          return {
            properties: {
              //always force the Relevant dropdown and show all entries
              'relevant-force': true,
              'relevant-filter': false,
              datalist: datalist
            },
            formFactor: 'mini'
          };
        });
    };
    return builder;
  };

  /**
   * Returns the `range` facet set on the editor, only if it is present and
   * is of type `baja:EnumRange`.
   *
   * @private
   * @returns {baja.EnumRange} range facet set, or undefined if not present
   * or not `EnumRange`
   */
  DynamicEnumEditor.prototype.$getRangeFacet = function () {
    var range = this.properties().getValue('range');
    if (baja.hasType(range, 'baja:EnumRange')) { return range; }
  };

  /**
   * Reads the currently selected `<option>` value and returns a new
   * `DynamicEnum` using that value as the ordinal.
   *
   * @returns {Promise.<baja.DynamicEnum>} a `DynamicEnum` instance
   * with the selected ordinal
   */
  DynamicEnumEditor.prototype.doRead = function () {
    var that = this, en = that.value();

    return that.getBuilder().readAll()
      .spread(function (str) {
        if (!str) { return baja.DynamicEnum.DEFAULT; }

        var ordinal = parseInt(str, 10);

        if (isNaN(ordinal)) {
          throw new Error('invalid ordinal: ' + str);
        } else {
          return baja.DynamicEnum.make({
            range: that.$getRangeFacet() || en.getRange(),
            ordinal: ordinal
          });
        }
      });
  };

  return (DynamicEnumEditor);
});
