/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/*jshint browser: true */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/IconEditor
 */
define(['baja!',
        'bajaux/icon/iconUtils',
        'underscore',
        'nmodule/webEditors/rc/fe/baja/BaseEditor',
        'nmodule/webEditors/rc/fe/baja/util/slotUtils',
        'css!nmodule/icons/sprite/sprite'], function (
         baja,
         iconUtils,
         _,
         BaseEditor,
         slotUtils) {

  'use strict';

  var getSlotIcon = slotUtils.getSlotIcon,
      toHtml = iconUtils.toHtml;

////////////////////////////////////////////////////////////////
// Support functions
////////////////////////////////////////////////////////////////

  /**
   * Given an array of URI strings, convert them to a `baja.Icon` (converting
   * `/module/` to `module://` ORD syntax).
   *
   * @inner
   * @param {Array.<String>} uris
   * @returns {baja.Icon}
   */
  function urisToIcon(uris) {
    return baja.Icon.make(_.map(uris, function (uri) {
      return uri.replace('/module/', 'module://');
    }));
  }

  /**
   * Get/make the appropriate `baja.Icon` instance for the loaded value.
   *
   * @inner
   * @param {baja.Icon|baja.Value|baja.Slot|Array.<String>} value
   * @returns {baja.Icon}
   */
  function toIcon(value) {
    if (baja.hasType(value, 'baja:Icon')) {
      return value;
    }

    if (baja.hasType(value)) {
      return value.getType().getIcon();
    }

    if (value instanceof baja.Slot) {
      return urisToIcon(getSlotIcon(value));
    }

    if (_.isArray(value)) {
      return urisToIcon(value);
    }
  }

////////////////////////////////////////////////////////////////
// IconEditor
////////////////////////////////////////////////////////////////

  /**
   * A display-only editor for working with `Icon`s.
   *
   * It supports the following `bajaux` Properties:
   *
   * - `sync`: if `true`, will force the image to be fully downloaded before
   *   `load()` completes. Use this if you are relying on the icon image
   *   dimensions for layout purposes.
   *
   * @private
   * @class
   * @extends module:nmodule/webEditors/rc/fe/baja/BaseEditor
   * @alias module:nmodule/webEditors/rc/fe/baja/IconEditor
   */
  var IconEditor = function IconEditor() {
    BaseEditor.apply(this, arguments);
  };
  IconEditor.prototype = Object.create(BaseEditor.prototype);
  IconEditor.prototype.constructor = IconEditor;

  /**
   * Add `IconEditor` class.
   *
   * @param {JQuery} dom
   */
  IconEditor.prototype.doInitialize = function (dom) {
    dom.addClass('IconEditor');
    dom.html('<div class="wrapper"/>');
  };

  /**
   * Load in an icon for display. This function can take a value of several
   * different kinds:
   *
   * - a `baja.Icon` directly
   * - any other `baja.Value` - its Type's icon will be shown
   * - a `baja.Slot`: a Property will show that property's Type, while an
   *   Action or Topic will show a generic `action.png` or `topic.png` icon
   * - an array of image URIs, which will be converted directly
   *
   * @param {baja.Icon|baja.Value|baja.Slot|Array.<String>} icon
   */
  IconEditor.prototype.doLoad = function (icon) {
    var sync = this.properties().getValue('sync'),
        jq = this.jq().children('.wrapper');

    return toHtml(toIcon(icon), sync)
      .then(function (html) {
        jq.html(html);
      });
  };

  /**
   * Resolve the currently loaded icon. Note that if any other value than a
   * `baja.Icon` was loaded in `doLoad`, it will be converted to a `baja.Icon`
   * when read.
   *
   * @returns {baja.Icon}
   */
  IconEditor.prototype.doRead = function () {
    return toIcon(this.value());
  };

  /**
   * Remove `IconEditor` class.
   */
  IconEditor.prototype.doDestroy = function () {
    this.jq().removeClass('IconEditor');
  };

  return (IconEditor);
});
