/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/TypeConfigEditor
 */
define(['baja!',
        'log!nmodule.webEditors.rc.fe.baja.TypeConfigEditor',
        'bajaux/events',
        'bajaux/Properties',
        'jquery',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/baja/util/facetsUtils',
        'nmodule/webEditors/rc/fe/baja/util/typeUtils',
        'nmodule/webEditors/rc/util/Switchboard',
        'nmodule/webEditors/rc/wb/PropertySheet',
        'nmodule/webEditors/rc/wb/PropertySheetRow',
        'nmodule/webEditors/rc/servlets/typeConfig'], function (
         baja,
         log,
         events,
         Properties,
         $,
         Promise,
         _,
         facetsUtils,
         typeUtils,
         Switchboard,
         PropertySheet,
         PropertySheetRow,
         typeConfig) {

  'use strict';

  var getTypeConfigInfo = typeConfig.getTypeConfigInfo,
      getConfigurableInfo = typeConfig.getConfigurableInfo,

      toProperties = facetsUtils.toProperties,

      importTypes = typeUtils.importTypes,
      toTypeSpec = typeUtils.toTypeSpec,

      logError = log.severe.bind(log),

      MODIFY_EVENT = events.MODIFY_EVENT;

  var nullConfig = { config: [] };

  function getByKey(config, key) {
    for (var i = 0; i < config.length; i++) {
      if (config[i].key === key) {
        return config[i];
      }
    }
  }

  function isNull(typeSpec) {
    return !typeSpec || (String(typeSpec) === 'null');
  }

  function removeAllUnknown(comp, keys, batch) {
    return Promise.all(comp.getSlots().toArray().map(function (slot) {
      return keys.indexOf(slot.getName()) < 0 &&
        comp.remove({ slot: slot, batch: batch });
    }));
  }

  /**
   * Resolve true if the type spec represents a valid baja Type. Resolve false
   * if invalid or null.
   *
   * @param typeSpec
   * @returns {Promise}
   */
  function validateTypeSpec(typeSpec) {
    if (isNull(typeSpec)) {
      return Promise.resolve(false);
    }

    return importTypes(String(typeSpec))
      .then(function () {
        return true;
      })
      .catch(function () {
        return false;
      });
  }

  /**
   * Editor for working with `TypeConfig`s. As of 8/25/2014 it works just barely
   * well enough to handle web profiles in the user manager and as such should
   * be considered private, probably-broken API.
   *
   * It makes use of two station-side services:
   *
   * - `com.tridium.webeditors.servlets.TypeConfigServlet`: for retrieving
   *   information about `TypeConfig` and `IConfigurable` classes that are not
   *   supported in BajaScript directly.
   * - `com.tridium.webeditors.ssc.BTypeConfigSSCHandler': for saving changes.
   *   The `IConfigurable` data will be sent to the station packaged in a
   *   `BComponent` and applied to the `TypeConfig` using `syncFrom`.
   *   Theoretically, this should work with other `BTypeConfig`s that use the
   *   default implementation of `syncFrom`, by my confidence level is low.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/fe/baja/TypeConfigEditor
   * @extends module:nmodule/webEditors/rc/wb/PropertySheet
   */
  var TypeConfigEditor = function TypeConfigEditor(params) {
    PropertySheet.call(this, $.extend(true, {
      module: 'webEditors',
      keyName: 'TypeConfigEditor'
    }, params));

    new Switchboard(this)
      .allow('$getConfigForIConfigurable').oneAtATime()
      .allow('read').notWhile('$rebuild');
  };
  TypeConfigEditor.prototype = Object.create(PropertySheet.prototype);
  TypeConfigEditor.prototype.constructor = TypeConfigEditor;

  /**
   * The default type spec to use when the loaded TypeConfig does not have
   * a type spec specified (e.g. it's a brand new TypeConfig like
   * `baja.$('web:WebProfileConfig')`). Defaults to null, which will not work -
   * but will be retrieved from the servlet in `$preloadTypeConfigInfo`.
   *
   * @private
   * @field
   * @type baja.Value
   */
  TypeConfigEditor.prototype.$defaultType = toTypeSpec('null');

  /**
   * Get the currently-loaded target type (the `IConfigurable` type spec to
   * set on the `TypeConfig`'s `typeSpec` slot). If none has yet been loaded,
   * fall back to an existing default type spec.
   *
   * @private
   * @param {baja.Simple} defaultType the `baja:TypeSpec` to fall back to
   * @returns {Promise} promise to be resolved with the target type
   * spec
   */
  TypeConfigEditor.prototype.$getLoadedType = function (defaultType) {
    var that = this,
        ed = that.getBuilder().getEditorFor('typeSpec'),
        enteredValue = ed && !ed.$loading && ed.value();

    return validateTypeSpec(enteredValue)
      .then(function (valid) {
        return valid ? ed.read() : validateTypeSpec(defaultType)
          .then(function (valid) {
            return valid ? defaultType : that.$defaultType;
          });
      });
  };

  /**
   * Preload information about the loaded `TypeConfig` by calling out to the
   * `TypeConfigServlet`, and set as instance variables on the
   * `TypeConfigEditor`:
   *
   * - `$defaultType`: the default type to choose for the `typeSpec` slot, if
   *   none is currently set
   * - `$targetType`: restrict user choices for the `typeSpec` slot to only
   *   subclasses of this Type
   *
   * @private
   * @returns {Promise} promise to be resolved after info is preloaded
   */
  TypeConfigEditor.prototype.$preloadTypeConfigInfo = function () {
    var that = this;
    return getTypeConfigInfo(that.value().getType())
      .then(function (config) {
        that.$defaultType = toTypeSpec(config.defaultType);
        that.$targetType = config.targetType;
      });
  };

  /**
   * Retrieve information about the given key as reported by the currently
   * selected `IConfigurable`. Resolved object will have `key`, `displayName`,
   * `properties`, and `value` properties, suitable for passing to `makeFor`.
   *
   * @private
   * @param {String} key
   * @returns {Promise} promise to be resolved with a config object
   */
  TypeConfigEditor.prototype.$getConfigObject = function (key) {
    return this.$getConfigForIConfigurable()
      .then(function (obj) {
        return getByKey(obj.config, key);
      });
  };

  /**
   * Call out to the station, if necessary, to retrieve information about
   * the currently selected type spec. (Currently, this will blow up if the
   * selected type spec does not implement `IConfigurable`.)
   *
   * Afterwards, `$getConfigObject` for the given `IConfigurable` key will
   * resolve an object.
   *
   * @private
   * @returns {Promise} promise to be resolved with an object containing
   * all information about the currently selected `IConfigurable` type
   */
  TypeConfigEditor.prototype.$getConfigForIConfigurable = function () {
    var that = this,
        typeConfig = that.getBuilder().getDataSource();

    return that.$getLoadedType(typeConfig.get('typeSpec'))
      .then(function (configurableSpec) {
        if (isNull(configurableSpec)) {
          return nullConfig;
        }

        configurableSpec = String(configurableSpec);

        if (that.$configurableTypeSpec === configurableSpec) {
          return that.$configurableConfig;
        }

        that.$configurableTypeSpec = configurableSpec;

        return getConfigurableInfo(typeConfig.getType(), configurableSpec)
          .then(function (obj) {
            that.$configurableConfig = obj;
            return obj;
          });
      });
  };

  /**
   * Read currently entered values for all type config keys, then apply them
   * to the given TypeConfig. (This can be a new unmounted copy for a read, or
   * a mounted TypeConfig for a save.) Any keys currently on the TypeConfig,
   * that do not exist in the editor, will be removed from the TypeConfig.
   *
   * @private
   * @param {baja.Complex} config a `baja:TypeConfig`
   * @param {Boolean} [unmodify] set to true if child editors should be marked
   * as unmodified after applying values
   * @returns {Promise} promise to be resolved with the same `TypeConfig`
   * instance as passed in, with all keys written onto it
   */
  TypeConfigEditor.prototype.$applyKeysToConfig = function (config, unmodify) {
    var builder = this.getBuilder();

    return builder.getKeys()
      .then(function (keys) {
        var editors = _.map(keys, function (key) {
              return builder.getEditorFor(key);
            });

        return Promise.map(editors, function (editor) {
          return editor.read();
        })
          .then(function (values) {
            var batch = new baja.comm.Batch(),
                //don't use Promise.map here - need this to run synchronously
                //so the batch can be safely committed below
                sets = _.map(values, function (value, i) {
                  var key = keys[i],
                      has = config.has(key);
                  return config[has ? 'set' : 'add']({
                    slot: key,
                    value: value,
                    flags: has ? 0 : baja.Flags.READONLY,
                    batch: batch
                  })
                    .then(function () {
                      return unmodify && editors[i].setModified(false);
                    });
                }),
                prom = Promise.join(
                  Promise.all(sets),
                  removeAllUnknown(config, keys, batch)
                );

            batch.commit();
            return prom;
          });
      })
      .then(function () {
        return config;
      });
  };

  /**
   * Create the `CompositeBuilder` for the `TypeConfigEditor`. This builder
   * will assemble its sub-editors by calling out for information about the
   * currently loaded `TypeConfig` and `IConfigurable` classes, as necessary.
   *
   * @returns {module:nmodule/webEditors/rc/fe/config/ComplexCompositeBuilder}
   */
  TypeConfigEditor.prototype.makeBuilder = function () {
    var that = this,
        builder = PropertySheet.prototype.makeBuilder.apply(that, arguments),
        getConfigFor = builder.getConfigFor,
        getDisplayNameFor = builder.getDisplayNameFor,
        getIconFor = builder.getIconFor,
        $loadFor = builder.$loadFor;

    /**
     * All frozen slots on the `TypeConfig`, plus all config keys from the
     * currently selected `IConfigurable`.
     *
     * @returns {Promise} promise to be resolved with an array of keys
     */
    builder.getKeys = function () {
      var typeConfig = this.getDataSource();

      return that.$getConfigForIConfigurable()
        .then(function (obj) {
          var frozenNames = _.map(typeConfig.getSlots().frozen().toArray(), String),
              configNames = _.map(obj.config, _.property('key'));
          return frozenNames.concat(configNames);
        });
    };

    /**
     * For slots from the `TypeConfig`, use the typical `PropertySheet` editor
     * config. Since `PropertySheet` doesn't deal with arbitrary keys from
     * `IConfigurable`s, configure these editors using information from
     * `TypeConfigServlet`.
     *
     * @param {String} key
     * @returns {Promise} promise to be resolved with a config object
     */
    builder.getConfigFor = function (key) {
      var typeConfig = this.getDataSource();

      if (typeConfig.has(key) && typeConfig.getSlot(key).isFrozen()) {
        return getConfigFor.apply(this, arguments);
      }

      return that.$getConfigObject(key)
        .then(function (obj) {
          return $.extend({}, _.omit(obj, 'facets'), {
            type: PropertySheetRow,
            data: { builder: builder, key: key },
            properties: Properties.extend(toProperties(obj.facets), {
              alt: that.properties().getValue('alt')
            })
          });
        });
    };

    /**
     * Use the slot display name if the `TypeConfig` has that slot; otherwise
     * use the display name from the `TypeConfigServlet`.
     *
     * @param {String} key
     * @returns {Promise} promise to be resolved with a display name
     */
    builder.getDisplayNameFor = function (key) {
      var typeConfig = this.getDataSource();

      if (typeConfig.has(key)) {
        return getDisplayNameFor.apply(this, arguments);
      }

      return that.$getConfigObject(key)
        .then(function (obj) {
          return obj && obj.displayName;
        });
    };

    /**
     * Use the slot icon if the `TypeConfig` has that slot; otherwise use the
     * icon from the default value provided by the `TypeConfigServlet`.
     *
     * @param {String} key
     * @returns {Promise} promise to be resolved with an icon URI
     */
    builder.getIconFor = function (key) {
      var typeConfig = this.getDataSource();

      if (typeConfig.has(key)) {
        return getIconFor.apply(this, arguments);
      }

      return that.$getConfigObject(key)
        .then(function (obj) {
          return obj && obj.value.getType().getIcon().getImageUris()[0];
        });
    };

    /**
     * Use the slot value if the `TypeConfig` has that slot; otherwise use the
     * default value from the `TypeConfigServlet`.
     *
     * @param {String} key
     * @returns {Promise} promise to be resolved with a value to load
     */
    builder.getValueFor = function (key) {
      var typeConfig = this.getDataSource();

      if (typeConfig.has(key)) {
        return Promise.resolve(typeConfig.get(key));
      }

      return that.$getConfigObject(key)
        .then(function (obj) {
          return obj && obj.value;
        });
    };

    // TODO: why does BTypeConfig need to do this in the first place?
    // TODO: perhaps a "don't overwrite modified editors" flag to CompositeBuilder
    /**
     * `BTypeConfig#syncFrom` ensures that all values applied from the
     * `IConfigurable` get the READONLY flag. `BTypeConfigFE` then ignores this
     * flag and makes everything editable.
     *
     * This is dumb. But ensure that these editors get flipped from readonly
     * (from the flag) back to writable.
     *
     * @private
     * @param key
     * @returns {*}
     */
    builder.$loadFor = function (key) {
      var ed = this.getEditorFor(key);

      //don't touch modified editors
      return Promise.resolve(!ed.isModified() && $loadFor.apply(this, arguments))
        .then(function () {
          return ed.setReadonly(that.isReadonly());
        });
    };

    return builder;
  };

  // TODO: TypeSpecEditor could be condensed to a single dropdown
  /**
   * All slots should be shown and edited. The `typeSpec` slot will get the
   * appropriate faces to limit selectable types to the target type.
   *
   * @returns {Function}
   */
  TypeConfigEditor.prototype.getSlotFilter = function () {
    var that = this;

    return function (slot) {
      if (slot.getName() === 'typeSpec') {
        return {
          properties: {
            targetType: that.$targetType,
            showAbstract: false,
            showInterface: false
          }
        };
      }

      return true;
    };
  };

  /**
   * Rebuild editors for all keys. Split this out so we can use Switchboard
   * to prevent it from happening in the middle of a read and vice versa.
   *
   * @private
   * @returns {Promise}
   */
  TypeConfigEditor.prototype.$rebuild = function () {
    return this.getBuilder().buildAll();
  };

  /**
   * Ensure that all slots and config keys are shown and made editable whenever
   * a new `IConfigurable` type spec is chosen.
   *
   * @param {JQuery} dom
   * @returns {Promise}
   */
  TypeConfigEditor.prototype.doInitialize = function (dom) {
    var that = this;

    dom.on(MODIFY_EVENT, '.slot-typeSpec', function () {
      that.$rebuild().catch(logError);
    });

    return PropertySheet.prototype.doInitialize.apply(that, arguments);
  };

  /**
   * Load editors for all slots and config keys for the given `TypeConfig`.
   *
   * @param {baja.Component} value the `baja:TypeConfig` to load
   * @returns {Promise}
   */
  TypeConfigEditor.prototype.doLoad = function (value) {
    var that = this;

    return that.$preloadTypeConfigInfo()
      .then(function () {
        return PropertySheet.prototype.doLoad.call(that, value);
      })
      .then(function () {
        var typeSpecEd = that.getBuilder().getEditorFor('typeSpec');
        return validateTypeSpec(typeSpecEd.value())
          .then(function (valid) {
            if (!valid) {
              return typeSpecEd.load(toTypeSpec(that.$defaultType));
            }
          });
      });
  };

  /**
   * Resolve a new `TypeConfig` instance of the same type as the one currently
   * loaded, with all current keys applied to it.
   *
   * @returns {Promise}
   */
  TypeConfigEditor.prototype.doRead = function () {
    var config = baja.$(this.value().getType());
    return this.$applyKeysToConfig(config);
  };

  /**
   * Apply entered type config values to the loaded TypeConfig.
   *
   * @returns {Promise}
   */
  TypeConfigEditor.prototype.doSave = function () {
    return this.$applyKeysToConfig(this.value(), true);
  };

  return (TypeConfigEditor);
});
