/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/compat/CompositeAbsTimeEditor
 */
define(['baja!',
        'bajaScript/baja/obj/dateTimeUtil',
        'jquery',
        'Promise',
        'bajaux/events',
        'nmodule/webEditors/rc/fe/baja/BaseEditor'], function (
         baja,
         dateTimeUtil,
         $,
         Promise,
         events,
         BaseEditor) {

  'use strict';

  var MODIFY_EVENT = events.MODIFY_EVENT,
      getDateOnlyFormat = dateTimeUtil.getDateOnlyFormat,
      DEFAULT_TIME_FORMAT = dateTimeUtil.DEFAULT_TIME_FORMAT;

  /**
   * Editor for working with `baja:AbsTime` values. Splits the date and time
   * parts of the `AbsTime` into two separate field editors, using whichever
   * default editor is registered on those two types.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/fe/baja/compat/CompositeAbsTimeEditor
   * @extends module:nmodule/webEditors/rc/fe/baja/BaseEditor
   */
  var CompositeAbsTimeEditor = function CompositeAbsTimeEditor() {
    BaseEditor.apply(this, arguments);
  };
  CompositeAbsTimeEditor.prototype = Object.create(BaseEditor.prototype);
  CompositeAbsTimeEditor.prototype.constructor = CompositeAbsTimeEditor;

  /**
   * Get the field editor for the date part.
   *
   * @private
   * @returns {module:nmodule/webEditors/rc/fe/baja/BaseEditor}
   */
  CompositeAbsTimeEditor.prototype.$getDateEditor = function () {
    return this.jq().children('.date').data('widget');
  };

  /**
   * Get the field editor for the time part.
   *
   * @private
   * @returns {module:nmodule/webEditors/rc/fe/baja/BaseEditor}
   */
  CompositeAbsTimeEditor.prototype.$getTimeEditor = function () {
    return this.jq().children('.time').data('widget');
  };

  /**
   * Get the "null" label.
   *
   * @private
   * @returns {jQuery}
   */
  CompositeAbsTimeEditor.prototype.$getNullLabel = function () {
    return this.jq().children('label');
  };

  /**
   * When a new value is loaded or the readonly status changes, update the DOM
   * so that it reads "null" if `baja.AbsTime.DEFAULT` is loaded and the editor
   * is readonly.
   */
  CompositeAbsTimeEditor.prototype.$updateNullReadonly = function() {
    var that = this,
        nullReadonly = that.isReadonly() &&
          baja.AbsTime.DEFAULT.equals(that.value());

    if (nullReadonly) {
      that.jq().children('.date, .time').css('display', 'none');
      that.$getNullLabel().css('display', '');
    }
  };

  /**
   * Initialize two field editors for the date and time part of the `AbsTime`.
   *
   * @param {JQuery} dom
   * @returns {Promise}
   */
  CompositeAbsTimeEditor.prototype.doInitialize = function (dom) {
    var that = this,
        props = that.properties(),
        showDate = props.getValue('showDate') !== false &&
                   !!getDateOnlyFormat(props.getValue('timeFormat') ||
                                       DEFAULT_TIME_FORMAT),
        showTime = props.getValue('showTime') !== false;

    // unlike CompositeTimeEditor, CompositeAbsTimeEditor's default is to
    // not show the seconds editor
    if (!props.getValue('showSeconds') && !props.getValue('showMilliseconds')) {
      props.add('showSeconds', false);
    }

    dom.on(MODIFY_EVENT, '.editor', function () {
      that.setModified(true);
      return false;
    });

    dom.append('<label style="display:none;">null</label>');

    return Promise.join(
      that.makeChildFor({ value: baja.Date.DEFAULT, properties: props, formFactor: 'mini' }),
      that.makeChildFor({ value: baja.Time.DEFAULT, properties: props, formFactor: 'mini' })
    )
      .spread(function (dateEd, timeEd) {
        var dateDiv = $('<div class="date"/>').appendTo(dom),
            timeDiv = $('<div class="time"/>').appendTo(dom);

        if (!showDate) { dateDiv.hide(); }
        if (!showTime) { timeDiv.hide(); }

        return Promise.join(
          dateEd.initialize(dateDiv), timeEd.initialize(timeDiv)
        );
      });
  };

  /**
   * Load the date and time part of the `AbsTime` into the two field editors.
   *
   * @param {baja.AbsTime} value
   * @returns {Promise}
   */
  CompositeAbsTimeEditor.prototype.doLoad = function (value) {
    this.$updateNullReadonly();
    return Promise.join(
      this.$getDateEditor().load(value.getDate()),
      this.$getTimeEditor().load(value.getTime())
    );
  };

  /**
   * Read the date and time editors and assemble their values into an `AbsTime`
   * instance.
   *
   * @returns {Promise} promise to be resolved with an `AbsTime`
   */
  CompositeAbsTimeEditor.prototype.doRead = function () {
    var value = this.value(),
        offset = value && value.getOffset() || 0;

    return Promise.join(
      this.$getDateEditor().read(),
      this.$getTimeEditor().read()
    )
      .spread(function (date, time) {
        return baja.AbsTime.make({
          date: date,
          time: time,
          offset: offset
        });
      });
  };

  /**
   * Set child editors readonly/writable.
   *
   * @param {Boolean} readonly
   * @returns {Promise}
   */
  CompositeAbsTimeEditor.prototype.doReadonly = function (readonly) {
    this.$updateNullReadonly();
    return this.getChildEditors().setAllReadonly(readonly);
  };

  /**
   * Set child editors enabled/disabled.
   *
   * @param {Boolean} enabled
   * @returns {Promise}
   */
  CompositeAbsTimeEditor.prototype.doEnabled = function (enabled) {
    return this.getChildEditors().setAllEnabled(enabled);
  };

  /**
   * Destroy the child editors.
   *
   * @returns {Promise}
   */
  CompositeAbsTimeEditor.prototype.doDestroy = function () {
    return this.getChildEditors().destroyAll();
  };

  return (CompositeAbsTimeEditor);
});
