/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/baja/util/Attachable
 */
define(['underscore',
        'nmodule/webEditors/rc/fe/baja/util/compUtils',
        'nmodule/webEditors/rc/fe/baja/util/slotUtils',
        'nmodule/webEditors/rc/fe/baja/util/typeUtils'], function (
         _,
         compUtils,
         slotUtils,
         typeUtils) {
  'use strict';

  var slotPathFromAncestor = compUtils.slotPathFromAncestor,
    globMatch = slotUtils.globMatch,
    isComponent = typeUtils.isComponent;

  /**
   * Utility class for managing Baja event subscription and unsubscription for
   * a component and its children.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/fe/baja/util/Attachable
   * @param {baja.Component} comp the component onto which to attach event
   * handlers
   */
  var Attachable = function Attachable(comp) {
    if (!isComponent(comp)) {
      throw new Error('component required');
    }

    var that = this;

    that.$handlers = {};
    that.$comp = comp;

    var addedHandler = that.$addedHandler = function (prop) {
      var newComp = this.get(prop);

      if (isComponent(newComp)) {
        _.each(that.$handlers, function (handlerArray, eventPath) {
          if (!handlerArray) {
            return;
          }

          var targets = that.$getEventTargets(eventPath, comp);
          _.each(targets, function (target) {
            if (target.component === newComp) {
              _.each(handlerArray, function (handler) {
                newComp.attach(target.event, handler);
              });
            }
          });
        });
      }
    };

    comp.attach('added', addedHandler);
  };

  /**
   * Find all nested targets for this Baja event. Supports regex syntax.
   *
   * @private
   * @param {String} event Baja event name, or slot path-like syntax
   * @param {baja.Component} comp the root component of the Attachable
   * @returns {Array.<Object>}
   */
  Attachable.prototype.$getEventTargets = function (event, comp) {
    if (event.match(/\*\*/)) {
      return getEventTargetsRecursive(event, comp);
    }
    var that = this,
        slotPatterns = event.split('/'),
        eventName = slotPatterns.pop(),
        descendants = slotPatterns.slice(1).concat([eventName]).join('/');

    function toTarget(comp) {
      return { component: comp, event: eventName };
    }

    if (slotPatterns.length) {
      var regex = new RegExp('^' + slotPatterns[0]),
          kids = comp.getSlots().is('baja:Component').filter(function (slot) {
            return !!slot.getName().match(regex);
          }).toValueArray(),
          kidTargets = _.map(kids, function (kid) {
            return that.$getEventTargets(descendants, kid);
          });

      return _.flatten(kidTargets);
    }

    return [ toTarget(comp) ];
  };

  function getEventTargetsRecursive(event, comp) {
    var slotPatterns = event.split('/'),
      eventName = slotPatterns.pop(),
      slotPattern = slotPatterns.join('/');
    return _.flatten(getAllKids(comp))
      .filter(function (kid) {
        return globMatch(slotPathFromAncestor(comp, kid).join('/'), slotPattern);
      })
      .map(function (component) {
        return { component: component, event: eventName };
      });
  }

  function getAllKids(comp) {
    return comp.getSlots().is('baja:Component').toValueArray()
      .map(function (kid) { return [ kid ].concat(getAllKids(kid)); });
  }

  /**
   * Attach event handlers for a specified set of Baja events.
   *
   * Event keys can be straight Baja event names, in which they will be attached
   * to the main component passed to the constructor. They can also attach
   * handlers to child components using a slot path-like syntax; see examples.
   *
   * When attaching to child components, remember that no Baja event handlers
   * will be fired unless those children are subscribed. Either subscribe them
   * manually or use a `DepthSubscriber`.
   *
   * @param {Object} obj a map from event strings to handler functions
   *
   * @example
   *   <caption>Attach handlers directly to a component.</caption>
   *
   *   resolveAndSubscribe('station:|slot:/someComponent')
   *     .then(function (comp) {
   *       var att = new Attachable(comp);
   *       att.attach({
   *         added: function (prop) {
   *           baja.outln(prop + ' added');
   *         },
   *         removed: function (prop) {
   *           baja.outln(prop + ' removed');
   *         }
   *       });
   *
   *       //then later...
   *
   *       att.detach();
   *     });
   *
   * @example
   *   <caption>Attach handlers to a component and also to specified child
   *   components.</caption>
   *
   *   resolveAndSubscribeDeep('station:|slot:/Services', 2)
   *     .then(function (services) {
   *       var att = new Attachable(services);
   *       att.attach({
   *         added: function (prop) {
   *           baja.outln(prop + ' service added');
   *         },
   *         'UserService/added': function (prop) {
   *           baja.outln(prop + ' user added');
   *         }
   *       });
   *     });
   *
   * @example
   *   <caption>Attach handlers to multiple nested children of the component,
   *   using a regex syntax.</caption>
   *
   *   resolveAndSubscribeDeep('station:|slot:/Services/UserService', 2)
   *     .then(function (userService) {
   *       var att = new Attachable(userService);
   *       att.attach({
   *         '.+/changed': function (prop) {
   *           if (prop.getName() === 'fullName') {
   *             baja.outln('user ' + this.getName() + ' changed full name');
   *           }
   *         }
   *       });
   *     });
   */
  Attachable.prototype.attach = function (obj) {
    var that = this,
        comp = that.$comp,
        handlers = that.$handlers,
        addedHandler = that.$addedHandler;

    _.each(obj, function (handler, event) {
      var arr = handlers[event] || (handlers[event] = []),
          targets = that.$getEventTargets(event, comp);

      arr.push(handler);

      _.each(targets, function (target) {
        target.component.attach(target.event, handler);
        target.component.attach('added', addedHandler);
      });
    });
  };

  /**
   * Detach all Baja event handlers previously attached using `attach()`.
   */
  Attachable.prototype.detach = function () {
    var that = this,
        comp = that.$comp,
        handlers = that.$handlers,
        addedHandler = that.$addedHandler;

    comp.detach('added', addedHandler);

    _.each(handlers, function (arr, event) {
      _.each(arr, function (handler) {
        var targets = that.$getEventTargets(event, comp);

        _.each(targets, function (target) {
          target.component.detach(target.event, handler);
          target.component.detach('added', addedHandler);
        });
      });
    });
  };

  return Attachable;
});
