/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

define(['baja!',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/baja/util/numberUtils'], function (
         baja,
         Promise,
         _,
         numberUtils) {
  'use strict';

  var toDisplay = numberUtils.toDisplay;

  /**
   * @private
   * @exports nmodule/webEditors/rc/fe/baja/util/rangeUtils
   */
  var exports = {};

  /**
   * Gets min/max info from the facets, and a string representation
   * of that range in the form `[min - max]`. `min` and `max` default to
   * `-inf` and `+inf` if not present.
   *
   * `min` and `max` will be converted to number with `valueOf`, so `RelTime`s
   * etc. work as well.
   *
   * @param {Object|baja.Facets} facets
   * @returns {Promise.<{ min: number, max: number, rangeStr: string }>}
   */
  exports.getRangeInfo = function (facets) {
    if (!facets || typeof facets !== 'object') {
      return Promise.reject(new Error('facets required'));
    }

    var obj = toObject(facets),
        max = obj.max,
        min = obj.min,
        rangeStr = '';

    if (!min && min !== 0) {
      min = baja.Double.NEGATIVE_INFINITY;
    }

    if (!max && max !== 0) {
      max = baja.Double.POSITIVE_INFINITY;
    }

    if (!((min === baja.Float.NEGATIVE_INFINITY || min === baja.Double.NEGATIVE_INFINITY) &&
          (max === baja.Float.POSITIVE_INFINITY || max === baja.Double.POSITIVE_INFINITY))) {
      return Promise.all([ toDisplay(min, obj), toDisplay(max, obj) ])
        .spread(function (minDisplay, maxDisplay) {
          return {
            min: min.valueOf(),
            max: max.valueOf(),
            rangeStr: '[' + minDisplay + ' - ' + maxDisplay + ']'
          };
        });
    }

    return Promise.resolve({
      min: min.valueOf(),
      max: max.valueOf(),
      rangeStr: rangeStr
    });
  };

  /**
   * Checks to see if the given value is in range according to any `min` or
   * `max` values set on the given facets. Throws an appropriate message if
   * out of range, otherwise returns nothing.
   *
   * @param {Number} value number to check, or any value that can convert to
   * a number with `valueOf`
   * @param {Object|baja.Facets} facets
   * @returns {Promise} promise to be rejected if value is out of range
   */
  exports.inRange = function (value, facets) {
    var obj = toObject(facets),
      min = obj.min,
      max = obj.max,
      num = value.valueOf(),
      tmpFacets = _.extend({}, obj);

    if (baja.hasType(value, 'baja:Integer') || baja.hasType(value, 'baja:Long')) {
      if (min === baja.Float.NEGATIVE_INFINITY || min === baja.Double.NEGATIVE_INFINITY) {
        tmpFacets.min = baja.Long.MIN_VALUE;
      }
      if (max === baja.Float.POSITIVE_INFINITY || max === baja.Double.POSITIVE_INFINITY) {
        tmpFacets.max = baja.Long.MAX_VALUE;
      }
    }

    return exports.getRangeInfo(tmpFacets)
      .then(function (rangeInfo) {
        var rangeStr = rangeInfo.rangeStr;

        if (max !== undefined && max !== null && num > max) {
          return Promise.all([ toDisplay(value, obj), toDisplay(max, obj) ])
            .spread(function (valueDisplay, maxDisplay) {
              throw new Error(valueDisplay + ' > ' + maxDisplay + ' ' + rangeStr);
            });
        } else if (min !== undefined && min !== null && num < min) {
          return Promise.all([ toDisplay(value, obj), toDisplay(min, obj) ])
            .spread(function (valueDisplay, minDisplay) {
              throw new Error(valueDisplay + ' < ' + minDisplay + ' ' + rangeStr);
            });
        }
      });
  };

  /**
   * Given an ordinal and an `EnumRange`, resolve the display string for the
   * associated tag.
   *
   * If the range has a `lexicon` facet specified, resolve that lexicon and look
   * for an entry matching the tag. Otherwise, return the tag itself (unescaped
   * according to `SlotPath` rules).
   *
   * @param {Number} ordinal
   * @param {baja.EnumRange} range
   * @returns {Promise} promise to be resolved with a display string for the
   * tag denoted by the given ordinal
   */
  exports.getEnumRangeDisplay = function (ordinal, range) {
    return baja.DynamicEnum.make(ordinal).toString({ range: range });
  };

  function toObject(facets) {
    return baja.hasType(facets, 'baja:Facets') ? facets.toObject() : facets;
  }

  return (exports);
});
