/*
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/registry/Registry
 */
define(['Promise',
        'underscore',
        'nmodule/js/rc/tinyevents/tinyevents',
        'nmodule/webEditors/rc/fe/registry/RegistryEntry',
        'nmodule/webEditors/rc/fe/registry/impl/JsRegistry'], function (
         Promise,
         _,
         tinyevents,
         RegistryEntry,
         JsRegistry) {

  'use strict';

  //noinspection JSAccessibilityCheck
  var $hashParams = RegistryEntry.$hashParams;

////////////////////////////////////////////////////////////////
// Support functions
////////////////////////////////////////////////////////////////

  function hashTypeSpecAndParams(typeSpec, params) {
    return typeSpec + ',' + $hashParams(params);
  }

  function resolveIt(entry) { return entry.resolve(); }
  function resolveThemAll(entries) { return Promise.all(_.map(entries, resolveIt)); }

  /**
   * Resolve the first entry in the list. If it fails to resolve, walk down
   * the list until we find one that does. If all fail to resolve, just resolve
   * falsy.
   *
   * @inner
   * @param {Array.<module:nmodule/webEditors/rc/fe/registry/RegistryEntry>} entries
   * @returns {Promise}
   */
  function resolveFirst(entries) {
    return Promise.resolve(entries.length && entries[0].resolve())
      .catch(function () {
        return resolveFirst(entries.slice(1));
      });
  }



////////////////////////////////////////////////////////////////
// Exports
////////////////////////////////////////////////////////////////

  /**
   * Query parameters to narrow the results resolved for a given registry key.
   *
   * @typedef {Object} module:nmodule/webEditors/rc/fe/registry/Registry~QueryParams
   * @property {String} [rjs] If given, only resolve entries that have this
   * particular RequireJS ID.
   * @property {Array.<String>} [deps] If given, only resolve entries that have
   * this exact set of RequireJS dependencies.
   * @property {Array.<String>} [hasAny] If given, only resolve entries that
   * have at least one of these `tags`.
   * @property {Array.<String>} [hasAll] If given, only resolve entries that
   * have all of these `tags`.
   */

  /**
   * Base class for Registry implementations (local JS registration,
   * agent-based, etc).
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/fe/registry/Registry
   * @mixes tinyevents
   * @param {Object} [obj] a JSON object to use to initially build this registry
   * (typically will be used to reconstitute a registry using the previous
   * output of `toJSON`). If omitted, registry will be empty on creation.
   */
  var Registry = function Registry(obj) {
    this.$memoize();
    tinyevents(this);
    this.$localReg = new JsRegistry((obj || {}).localReg);
  };

  /**
   * @returns {module:nmodule/webEditors/rc/fe/registry/impl/JsRegistry}
   */
  Registry.prototype.getLocal = function () {
    return this.$localReg;
  };

  /**
   * Register a RequireJS module ID locally for the given key. Upon
   * completion, will cause the registry to emit a `changed` tinyevent.
   *
   * @param {String} key
   * @param {module:nmodule/webEditors/rc/fe/registry/Registry~QueryParams} [params]
   * query parameters/metadata for this RequireJS module
   * @returns {Promise} promise to be resolved after registration is complete
   */
   Registry.prototype.register = function (key, params) {
     var that = this;
     return Promise.try(function () {
       that.$localReg.register(key, params);
       //local registrations can happen after querying the registry, so we must
       //clear memoization caches to ensure the registration takes effect.
       that.$clearMemoization();
       that.emit('changed');
     });
  };

  /**
   * By default, just queries from the entries registered locally. Most likely,
   * subclasses will override this with something more useful.
   *
   * @param {String} key
   * @param {module:nmodule/webEditors/rc/fe/registry/Registry~QueryParams} params
   * @returns {Promise.<Array.<module:nmodule/webEditors/rc/fe/registry/RegistryEntry>>}
   * promise to be resolved with an array of all matching `RegistryEntry`s
   */
  Registry.prototype.queryAll = function (key, params) {
    return this.$localReg.queryAll(key, params);
  };

  /**
   * By default, just queries from the entries registered locally. Most likely,
   * subclasses will override this with something more useful.

   * @param {String} key
   * @param {module:nmodule/webEditors/rc/fe/registry/Registry~QueryParams} params
   * @returns {Promise.<module:nmodule/webEditors/rc/fe/registry/RegistryEntry>}
   * promise to be resolved with the first matching`RegistryEntry`
   */
  Registry.prototype.queryFirst = function (key, params) {
    return this.$localReg.queryFirst(key, params);
  };

  /**
   * Perform a query on the registry and resolve all RequireJS modules
   * represented. This will resolve an array of Widget constructors, menu
   * agent functions, etc.
   *
   * @param {String} key
   * @param {module:nmodule/webEditors/rc/fe/registry/Registry~QueryParams} params
   * @returns {Promise.<Array.<*>>} promise to be resolved with an array of the
   * exported results of all RequireJS modules represented, or empty if none
   * found; rejected if the station could not be successfully queried for
   * registry info, or if any of the RequireJS modules failed to resolve
   */
  Registry.prototype.resolveAll = function (key, params) {
    return this.queryAll(key, params).then(resolveThemAll);
  };

  /**
   * Perform a query on the registry and attempt to resolve the first
   * matching entry's RequireJS module. Note that this differs from `resolveAll`
   * in that if the first entry fails to resolve (for instance, an invalid
   * RequireJS module ID), it will move on to the next entry and keep trying
   * to resolve all the way down until it can resolve *something*.
   *
   * @param {String} key
   * @param {module:nmodule/webEditors/rc/fe/registry/Registry~QueryParams} params
   * @returns {Promise.<*>} promise to be resolved with the exported
   * results of the first matching entry that successfully resolves its
   * RequireJS module ID, or undefined if none found
   */
  Registry.prototype.resolveFirst = function (key, params) {
    var that = this;

    return that.queryFirst(key, params)
      .then(function (entry) {
        return entry && entry.resolve()
          .catch(function () {
            return that.queryAll(key, params).then(resolveFirst);
          });
      });
  };

  /**
   * Return an object suitable for serialization using `JSON.stringify` or
   * similar. The returned object can be passed right back to a
   * `Registry` constructor to reconstitute later.
   *
   * @returns {Object}
   */
  Registry.prototype.toJSON = function () {
    return {
      localReg: this.$localReg.toJSON()
    };
  };

  /**
   * Set up memoizing on all registry query calls. Since registry contents
   * do not change at runtime, this should be safe to call.
   * @private
   */
  Registry.prototype.$memoize = function () {
    var that = this;
    that.queryAll = _.memoize(that.queryAll, hashTypeSpecAndParams);
    that.queryFirst = _.memoize(that.queryFirst, hashTypeSpecAndParams);
    that.resolveAll = _.memoize(that.resolveAll, hashTypeSpecAndParams);
    that.resolveFirst = _.memoize(that.resolveFirst, hashTypeSpecAndParams);
  };

  /**
   * Clear memoization caches.
   * @private
   */
  Registry.prototype.$clearMemoization = function () {
    var that = this;

    if (!that.queryAll.cache) {
      return; //memoization not done
    }
    that.queryAll.cache = {};
    that.queryFirst.cache = {};
    that.resolveAll.cache = {};
    that.resolveFirst.cache = {};
  };

  return Registry;
});
