/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/registry/RegistryEntry
 */
define(['underscore',
        'nmodule/js/rc/asyncUtils/asyncUtils'], function (
         _,
         asyncUtils) {

  'use strict';

  var doRequire = asyncUtils.doRequire;

////////////////////////////////////////////////////////////////
// Support functions
////////////////////////////////////////////////////////////////

  /**
   * Check to see if any of the desired values exist in the array.
   *
   * @inner
   * @param {Array} toTest
   * @param {Array} desired
   * @returns {boolean} true if `desired` is empty ("I don't care") or if any
   * of the values in `desired` also exist in `toTest`
   */
  function hasAny(toTest, desired) {
    if (!desired.length) {
      return true; //don't care
    }

    for (var i = 0; i < desired.length; i++) {
      if (toTest.indexOf(desired[i]) >= 0) {
        return true;
      }
    }
  }

  /**
   * Check to see if all of the desired values exist in the array.
   *
   * @inner
   * @param {Array} toTest
   * @param {Array} desired
   * @returns {boolean} true if all values in `desired` also exist in `toTest`
   * (independent of order)
   */
  function hasAll(toTest, desired) {
    return !_.difference(desired, toTest).length;
  }

////////////////////////////////////////////////////////////////
// Exports
////////////////////////////////////////////////////////////////

  /**
   * An entry intended to be stored in a `JsRegistry`, representing a single
   * RequireJS module. Additional parameters can be used to store metadata
   * about this module, allowing it to be queried later in more detail.
   *
   * There is usually no reason to instantiate this class directly;
   * `JsRegistry` will create them as needed.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/fe/registry/RegistryEntry
   * @param {module:nmodule/webEditors/rc/fe/registry/Registry~QueryParams} params
   */
  var RegistryEntry = function RegistryEntry(params) {
    if (!params || typeof params.rjs !== 'string') {
      throw new Error('rjs parameter required');
    }

    this.$rjs = params.rjs;
    this.$deps = params.deps || [];
    this.$tags = params.tags || [];
  };

  /**
   * Resolve the RequireJS module (and any dependencies) that is represented
   * by this module.
   *
   * @returns {Promise} promise to be resolved with the contents of
   * the RequireJS module, or rejected if the module ID or any of its
   * dependencies could not be loaded
   */
  RegistryEntry.prototype.resolve = function () {
    return doRequire(this.$rjs, this.$deps);
  };

  /**
   * Compute a hash string for a query params object. Since the registry
   * contents should never change while apps are running, we can memoize calls
   * to registry query functions.
   *
   * @private
   * @param {Object} params params object as expected by `matches()`
   * @returns {String} a hash string
   */
  RegistryEntry.$hashParams = function (params) {
    params = params || {};
    var tags = params.tags,
        all = params.hasAll,
        any = params.hasAny,
        rjs = params.rjs,
        deps = params.deps;

    return 't' + (tags ? tags.join() : '') +
      'r' + (rjs || '') +
      'd' + (deps ? deps.join() : '') +
      'l' + (all ? all.join() : '') +
      'y' + (any ? any.join() : '');
  };

  /**
   * Check to see if this entry's metadata matches a registry query.
   *
   * @param {module:nmodule/webEditors/rc/fe/registry/Registry~QueryParams|module:nmodule/webEditors/rc/fe/registry/RegistryEntry} [params]
   * @returns {boolean} true if this entry matches the query parameters and
   * so should be included in the results
   */
  RegistryEntry.prototype.matches = function (params) {
    if (params instanceof RegistryEntry) {
      //noinspection JSUnresolvedFunction
      params = params.toJSON();
    }

    params = params || {};

    var any = params.hasAny || params.tags,
        all = params.hasAll,
        rjs = params.rjs,
        deps = params.deps;

    if (any && !hasAny(this.$tags, any)) {
      return false;
    }

    if (all && !hasAll(this.$tags, all)) {
      return false;
    }

    if (rjs && this.$rjs !== rjs) {
      return false;
    }

    //noinspection RedundantIfStatementJS
    if (deps && !hasAll(this.$deps, deps)) {
      return false;
    }

    return true;
  };

  /**
   * Get the RequireJS ID for this entry.
   * @returns {string}
   */
  RegistryEntry.prototype.getJsId = function () { return this.$rjs; };

  /**
   * Get the list of RequireJS dependencies that must be loaded before requiring
   * the main RequireJS ID.
   * @returns {Array.<String>}
   */
  RegistryEntry.prototype.getJsDependencies = function () { return this.$deps; };

  /**
   * Get a list of tags that serve as metadata for this entry.
   * @returns {Array.<String>}
   */
  RegistryEntry.prototype.getTags = function () { return this.$tags; };

  /**
   * Return a raw object representation of this entry. As a contractual
   * requirement, the output of this function should be able to be passed right
   * back to the constructor to create a new instance. Return `undefined` to
   * indicate that this entry should never be serialized.
   * @returns {Object|undefined}
   */
  RegistryEntry.prototype.toJSON = function () {
    return {
      rjs: this.getJsId(),
      deps: this.getJsDependencies(),
      tags: this.getTags()
    };
  };

  return RegistryEntry;
});
