/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/registry/impl/AgentRegistry
 */
define(['Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/baja/util/typeUtils',
        'nmodule/webEditors/rc/fe/registry/Registry',
        'nmodule/webEditors/rc/fe/registry/servletViews'], function (
         Promise,
         _,
         typeUtils,
         Registry,
         servletViews) {

  'use strict';

  var getSuperTypeChain = typeUtils.getSuperTypeChain,
      importTypes = typeUtils.importTypes,
      FORM_FACTORS = [
        'web:IFormFactorMini',
        'web:IFormFactorCompact',
        'web:IFormFactorMax'
      ];

////////////////////////////////////////////////////////////////
// Support functions
////////////////////////////////////////////////////////////////

  function isFormFactor(t) { return _.contains(FORM_FACTORS, t); }

  function containsFormFactor(tags) { return !!_.find(tags, isFormFactor); }

////////////////////////////////////////////////////////////////
// Exports
////////////////////////////////////////////////////////////////

  /**
   * A registry for retrieving, storing, and querying information about
   * RequireJS modules registered on baja TypeSpecs. This information will be
   * retrieved directly from the station registry as `web:IJavaScript` classes
   * registered as agents on other Types.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/fe/registry/impl/AgentRegistry
   * @extends module:nmodule/webEditors/rc/fe/registry/Registry
   */
  var AgentRegistry = function AgentRegistry() {
    Registry.apply(this);
    this.$fetchedStatus = {};
  };
  AgentRegistry.prototype = Object.create(Registry.prototype);
  AgentRegistry.prototype.constructor = AgentRegistry;

  /**
   * Sanitize registration parameters. In particular, if registering with a
   * form factor interface like `web:IFormFactorMini`, ensure that the
   * `web:IFormFactor` superinterface is included so that the entry will be
   * returned on future queries for `web:IFormFactor`. This is necessary
   * because the `RegAgentOnTypeServlet` et al do not include superinterfaces,
   * only those explicitly declared on the class.
   *
   * @private
   * @param {Object} params registration params
   * @returns {Object} sanitized params
   */
  AgentRegistry.$processParamsForRegistration = function (params) {
    var tags = params && params.tags;
    if (tags && containsFormFactor(tags) && !_.contains(tags, 'web:IFormFactor')) {
      params.tags = tags.concat([ 'web:IFormFactor' ]);
    }
    return params;
  };

  /**
   * Check to see if information about the given type spec has already been
   * queried from the station.
   *
   * @private
   * @param {Type|String} typeSpec
   * @returns {Boolean}
   */
  AgentRegistry.prototype.$isFetched = function (typeSpec) {
    return this.$fetchedStatus[typeSpec];
  };

  /**
   * Mark the given type spec as having been queried from the station: no more
   * calls to the registry servlet will be performed for this type spec. (Note
   * that since we added registry info to BajaScript, this no longer saves any
   * actual network traffic; but prevents types from double-registering.)
   *
   * @private
   * @param {Type|String} typeSpec
   */
  AgentRegistry.prototype.$markFetched = function (typeSpec) {
    //TODO: should i be able to un-mark a type so it gets re-fetched?
    this.$fetchedStatus[typeSpec] = true;
  };

  /**
   * Perform the work of fetching all `web:IJavaScript` data down from the
   * station and storing it. After this is complete, the data can be queried
   * and used without ever having to query the station about this type spec
   * again.
   *
   * @private
   * @param {Type} type
   * @returns {Promise} promise to be resolved after all data has been fetched
   * and stored. No data is returned from this function - query the type spec
   * afterwards using the usual `queryAll` etc. If the type has already been
   * fetched, this promise will be immediately resolved without doing any work.
   */
  AgentRegistry.prototype.$pullFromStation = function(type) {
    var that = this,
        stationReg = that.getLocal(),
        toFetch;

    if (that.$isFetched(type)) { return Promise.resolve(); }

    toFetch = _.filter(getSuperTypeChain(type), function (superType) {
      return !that.$isFetched(superType);
    });

    return servletViews.getEntries(toFetch)
      .then(function (arraysOfEntries) {
        _.each(arraysOfEntries, function (entries, i) {
          var typeSpec = toFetch[i];
          if (that.$isFetched(typeSpec)) { return; }
          _.each(entries.reverse(), function (entry) {
            stationReg.register(typeSpec, AgentRegistry.$processParamsForRegistration({
              deps: entry.getJsDependencies(),
              tags: entry.getTags(),
              rjs: entry.getJsId()
            }));
            that.$markFetched(typeSpec);
          });
        });
      })
      .then(function () {
        that.emit('changed');
      });
  };

  /**
   * Prepare for a registry query by ensuring that the type is imported and
   * that we've retrieved and preloaded all necessary agent registration info
   * from the station.
   *
   * @private
   * @param {String} typeSpec
   * @returns {Promise} promise to be resolved, with the imported type,
   * after all info is preloaded
   */
  AgentRegistry.prototype.$preQuery = function (typeSpec) {
    var that = this;
    return importTypes(typeSpec)
      .spread(function (t) {
        return that.$pullFromStation(t)
          .then(function () {
            return t;
          });
      });
  };


  /**
   * Query all entries from the station that match the given type spec and
   * parameters. The results will be retrieved using the station's agent
   * registration information if needed. Note that results from supertypes
   * will not be included.
   *
   * @param {Type|String} typeSpec
   * @param {Object} params query parameters
   * @returns {Promise} promise to be resolved with an array of
   * `RegistryEntry` objects, or empty if none found; rejected if the station
   * could not be successfully queried for registry info
   */
  AgentRegistry.prototype.queryAll = function (typeSpec, params) {
    var that = this;

    return that.$preQuery(typeSpec)
      .then(function () {
        return that.getLocal().queryAll(typeSpec, params);
      });
  };

  /**
   * Perform a query on the registry and return only the first matching
   * `RegistryEntry`. (This would be the default view, default menu agent, etc.)
   *
   * @param {Type|String} typeSpec
   * @param {Object} params query parameters
   * @returns {Promise} promise to be resolved with a `RegistryEntry`
   * object, or `undefined` if none found; rejected if the station could not
   * be successfully queried for registry info
   */
  AgentRegistry.prototype.queryFirst = function (typeSpec, params) {
    var that = this;

    return that.$preQuery(typeSpec)
      .then(function () {
        return that.getLocal().queryFirst(typeSpec, params);
      });
  };

  return AgentRegistry;
});
