/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/fe/registry/impl/JsRegistry
 */
define(['underscore',
        'nmodule/webEditors/rc/fe/registry/RegistryEntry'], function (
         _,
         RegistryEntry) {

  'use strict';

////////////////////////////////////////////////////////////////
// Support functions
////////////////////////////////////////////////////////////////

  /**
   * Find all entries that match the given parameters.
   *
   * @inner
   * @param {Array.<module:nmodule/webEditors/rc/fe/registry/RegistryEntry>} entries
   * @param {Object} params
   * @returns {Array.<module:nmodule/webEditors/rc/fe/registry/RegistryEntry>}
   * array of matching entries
   */
  function findAllMatching(entries, params) {
    return _.filter(entries || [], function (entry) {
      return entry.matches(params);
    });
  }

  /**
   * Find the first matching entry in the array of entries. Will start looking
   * at the end first - so more recently registered entries come back first.
   *
   * @inner
   * @param {Array.<module:nmodule/webEditors/rc/fe/registry/RegistryEntry>} entries
   * @param {Object} params
   * @returns {module:nmodule/webEditors/rc/fe/registry/RegistryEntry|undefined}
   * matching entry, or undefined if none found
   */
  function findFirstMatching(entries, params) {
    if (!entries) {
      return;
    }

    for (var i = entries.length - 1; i >= 0; i--) {
      if (entries[i].matches(params)) {
        return entries[i];
      }
    }
  }

  /**
   * Reconstitutes a database object from a plain JSON object (output previously
   * from `toJSON()`.
   *
   * @inner
   * @param {Object} db raw JSON object
   * @returns {Object} reconstituted database object
   */
  function fromObj(db) {
    _.each(db, function (arr) {
      _.each(arr, function (obj, i) {
        arr[i] = new RegistryEntry(obj);
      });
    });
    return db;
  }

////////////////////////////////////////////////////////////////
// JsRegistry
////////////////////////////////////////////////////////////////

  /**
   * A simple registry for keeping a collection of JS information. This is
   * used to map different keys to collections of RequireJS modules. Each
   * entry gets tagged with different metadata so that you can map many
   * different RequireJS modules of different types to the same key, and query
   * them as needed.
   *
   * In a BajaScript environment, for instance, the key will be a Type Spec,
   * and a collection of RequireJS modules (e.g. Widgets, MenuAgents) will be
   * mapped to those type specs.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/fe/registry/impl/JsRegistry
   * @param {Object} [db] a JSON object to use to initially build this registry
   * (typically will be used to reconstitute a registry using the previous
   * output of `toJSON`). If omitted, registry will be empty on creation.
   */
  var JsRegistry = function JsRegistry(db) {
    this.$db = fromObj(db || {});
  };

  /**
   * Returns the backing database object.
   *
   * @private
   * @returns {Object}
   */
  JsRegistry.prototype.$getDatabase = function () {
    return this.$db;
  };

  /**
   * Register a RequireJS module, along with other optional metadata, under the
   * given key.
   *
   * @param {String} key
   * @param {Object|String} params object literal containing metadata. This may
   * be a string, which is equivalent to registering the RequireJS module ID
   * directly with no metadata at all.
   * @param {String} params.rjs the RequireJS module ID
   * @param {Array.<String>} [params.deps] Dependencies to require before
   * loading the RequireJS module
   */
  JsRegistry.prototype.register = function (key, params) {
    var db = this.$db;

    if (typeof params === 'string') {
      params = { rjs: params };
    }

    if (!key || !params) {
      throw new Error('key and params required');
    }

    if (!params.rjs) {
      throw new Error('RequireJS module id required');
    }

    var records = db[key] || (db[key] = []),
        existing = findAllMatching(records, params);

    if (!existing.length) {
      records.push(new RegistryEntry(params));
    }
  };

  /**
   * Find all entries registered under this key, that match the requested
   * params. The resolved array will be returned in insertion order: the oldest
   * entry will be at index 0.
   *
   * @param {String} key
   * @param {Object} [params] query parameters. These will be passed to the
   * `matches` function of each entry to determine whether it is included in
   * the resolved array.
   * @returns {Array.<module:nmodule/webEditors/rc/fe/registry/RegistryEntry>}
   * array of `RegistryEntry`s (empty if no matching entries found).
   */
  JsRegistry.prototype.queryAll = function (key, params) {
    return findAllMatching(this.$db[key], params);
  };

  /**
   * Find the first entry registered under this key that matches the requested
   * params. Entries registered later chronologically are considered first;
   * that is, this will return the most-recently-registered matching entry.
   *
   * @param key
   * @param params
   * @returns {*}
   */
  JsRegistry.prototype.queryFirst = function (key, params) {
    return findFirstMatching(this.$db[key], params);
  };

  /**
   * Return an object suitable for serialization using `JSON.stringify` or
   * similar. The returned object can be passed right back to a `JsRegistry`
   * constructor to reconstitute later.
   *
   * @returns {Object}
   */
  JsRegistry.prototype.toJSON = function () {
    var json = {};
    _.each(this.$db, function (arr, key) {
      json[key] = _.map(arr, function (entry) {
        return entry.toJSON();
      });
    });
    return json;
  };

  return JsRegistry;
});
