/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/*jshint browser: true */

/**
 * @private
 * @module nmodule/webEditors/rc/servlets/password
 */
define(['baja!',
        'lex!webEditors',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/baja/util/compUtils',
        'nmodule/webEditors/rc/fe/baja/util/spaceUtils',
        'nmodule/webEditors/rc/fe/baja/util/subscriberUtils',
        'nmodule/webEditors/rc/fe/baja/util/typeUtils',
        'nmodule/webEditors/rc/util/htmlUtils'], function (
         baja,
         lexs,
         Promise,
         _,
         compUtils,
         spaceUtils,
         subscriberUtils,
         typeUtils,
         htmlUtils) {

  'use strict';

  var HTTPS = /^https:$/i,

      RPC_TYPE_SPEC = 'web:ChangeUserPasswordRpc',

      webEditorsLex = lexs[0],

      getNavOrd = compUtils.getNavOrd,
      unescapeHtml = htmlUtils.unescapeHtml,
      resolveService = spaceUtils.resolveService,
      leaseWhenNeeded = subscriberUtils.leaseWhenNeeded,
      isComplex = typeUtils.isComplex,

      getWebService = leaseWhenNeeded(_.partial(resolveService, 'web:WebService'),
        1000);

  /** @param {String} err */
  function reject(err) { return Promise.reject(new Error(err)); }

  function changeUserPasswords(usernames, passwords, params) {
    return baja.rpc({
      typeSpec: RPC_TYPE_SPEC,
      method: 'changeUserPasswords',
      args: [ usernames, _.map(passwords, btoa) ],
      batch: params && params.batch
    })
      .catch(function (err) {
        baja.error(err);
        throw new Error(webEditorsLex.get('password.changeError'));
      });
  }

  function validatePasswords(schemes, usernames, passwords, params) {
    return baja.rpc({
      typeSpec: RPC_TYPE_SPEC,
      method: 'validatePasswords',
      args: [ usernames, _.map(passwords, btoa), schemes ],
      batch: params && params.batch
    })
      .catch(function(err){
        baja.error(err);
        throw new Error(unescapeHtml(err.message));
      });
  }

  function setPassword(pwd, pwdSlotName, baseOrd, params) {
    return baja.rpc({
      typeSpec: "web:PasswordRpc",
      method: 'setPassword',
      args: [pwd, pwdSlotName, baseOrd],
      batch: params && params.batch
    })
      .catch(function (err) {
        baja.error(err);
        throw new Error(unescapeHtml(err.message));
      });
  }

  /**
   * Check to see if the user service has a property set to explicitly allow
   * doing a dangerous password reset over HTTP.
   *
   * TODO: should we mark the field editor in red or something if so?
   *
   * @inner
   * @returns {Promise} promise to be resolved with a boolean
   */
  function isInsecureAllowed() {
    return getWebService()
      .then(function (webService) {
        return !webService.get('requireHttpsForPasswords');
      });
  }

  /**
   * Utilities relating to password changes. These include some client-side
   * security checks, but they should be treated as feel-good measures purely
   * for UI warnings and the like. Anything sent up from this module in a
   * non-HTTPS environment is subject to man in the middle attacks.
   *
   * This module communicates with the following servlets:
   *
   * - `com.tridium.web.servlets.ChangeUserPasswordServlet`
   *
   * @private
   * @exports nmodule/webEditors/rc/servlets/password
   */
  var exports = {};

  /**
   * Check to see if we are running in a secured environment. Note that
   * this is a "feel-good" measure: if we're in an insecure environment no
   * amount of JS checks will make it secure for real.
   *
   * @private
   * @returns {Promise}
   */
  exports.$isSecure = function () {
    if (HTTPS.exec(location.protocol) || baja.isOffline()) {
      return Promise.resolve(true);
    }
    return isInsecureAllowed();
  };

  function requireSecure() {
    //noinspection JSAccessibilityCheck
    return exports.$isSecure()
      .then(function (secure) {
        if (!secure) {
          return reject('insecure environment');
        }
      });
  }

  /**
   * Change the passwords for the given users.
   *
   * @param {Array.<String>} usernames array of usernames
   * @param {Array.<String>} passwords array of plaintext passwords
   * @param {baja.Complex} base a base Complex to use to resolve the
   * `WebService`
   * @param {Object} params
   * @param {baja.comm.Batch} [params.batch]
   * @returns {Promise} promise to be resolved if the passwords are
   * successfully changed. Rejects if the passwords cannot be changed, or if
   * called from a non-SSL environment (no POST to the station will occur).
   */
  exports.changePasswords = function (usernames, passwords, base, params) {
    var reserve = baja.comm.Batch.reserve(params);

    return requireSecure()
      .then(function () {
        if (!_.isArray(usernames) || !_.isArray(passwords)) {
          return reject('usernames and passwords required');
        }
        if (usernames.length !== passwords.length) {
          return reject('different numbers of usernames and passwords');
        }
        if (!isComplex(base)) {
          return reject('base required');
        }

        return reserve.commit(changeUserPasswords(usernames, passwords, {batch: reserve}));
      });
  };

  /**
   * Verify the given passwords against the specified authentication schemes.
   *
   * @param {Array.<String>} schemes array of authentication scheme names
   * @param {Array.<String>} usernames array of usernames
   * @param {Array.<String>} passwords array of plaintext passwords
   * @param {baja.Complex} base a base Complex to use to resolve the
   * `WebService`
   * @param {Object} params
   * @param {baja.comm.Batch} [params.batch]
   * @returns {Promise} promise to be resolved if the passwords are valid for
   * their corresponding authentication schemes (meet strength requirements
   * etc.). Rejects if any passwords are invalid, or if called from a non-SSL
   * environment (no POST to the station will occur). The message of the
   * rejection Error can be JSON-parsed as a username -> reason map.
   */
  exports.validatePasswords = function (schemes, usernames, passwords, base, params) {
    var reserve = baja.comm.Batch.reserve(params);

    return requireSecure()
      .then(function () {
        if (!(_.isArray(schemes) && _.isArray(passwords) && _.isArray(usernames))) {
          return reject('schemes, usernames and passwords required');
        }
        if (schemes.length !== passwords.length) {
          return reject('different numbers of schemes and passwords');
        }
        if (schemes.length !== usernames.length) {
          return reject('different numbers of schemes and usernames');
        }
        if (!isComplex(base)) {
          return reject('base required');
        }

        return reserve.commit(validatePasswords(schemes, usernames, passwords, {batch: reserve}));
      });
  };

  /**
   * Set the password property of component (resolved via baseOrd)
   * @param {String} pwd - The password to set
   * @param {baja.Slot|String} pwdSlot - The frozen slot to set the password on
   * @param {baja.Complex|baja.Ord} base - The base complex or ord
   * @param {Object} params
   * @param {baja.comm.Batch} [params.batch]
   * @returns {Promise} Promise to be resolved for successful password set
   * or reject if insecure
   */
  exports.setPassword = function (pwd, pwdSlot, base, params) {
    var reserve = baja.comm.Batch.reserve(params);

    return requireSecure()
      .then(function () {
        var ord;
        if (isComplex(base)) {
          ord = getNavOrd(base);
        } else if (baja.hasType(base, 'baja:Ord')) {
          ord = base;
        } else {
          return reject('base required');
        }
        return reserve.commit(setPassword(pwd, String(pwdSlot), String(ord), {batch: reserve}));
      });
  };

  return (exports);
});
