/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/servlets/typeConfig
 */
define([ 'baja!',
        'Promise',
        'underscore' ], function (
         baja,
         Promise,
         _) {

  'use strict';

  var RPC_TYPE_ORD = 'type:webEditors:TypeConfigRpc';

  var getTypeConfigInfo = _.chain(baja.rpc)
    .bind(baja)
    .partial(RPC_TYPE_ORD, 'getTypeConfigInfo')
    .memoize()
    .value();

  var getConfigurableInfo = _.chain(baja.rpc)
    .bind(baja)
    .partial(RPC_TYPE_ORD, 'getConfigurableInfo')
    .memoize(String.prototype.concat)
    .value();

  function decodeAsync(str) { return baja.bson.decodeAsync(JSON.parse(str)); }

  /**
   * Utilities for retrieving information about `BTypeConfig` and
   * `IConfigurable` Types from the station.
   *
   * This module communicates with the following RPC services:
   *
   * - `com.tridium.webeditors.servlets.BTypeConfigRpc`
   *
   * @private
   * @exports nmodule/webEditors/rc/servlets/typeConfig
   */
  var exports = {};

  /**
   * Retrieve information about the given `BTypeConfig` class. The resolved
   * object will contain the following properties:
   *
   * - `defaultType`: the default type to preselect, if none is set
   * - `targetType`: only types extending this type should be available for
   *   selection
   *
   * To illustrate, the info retrieved for `web:WebProfileConfig` would have
   * `hx:HTML5HxProfile` as the `defaultType` (that should be
   * preselected on a brand new web profile) and `targetType` would be
   * `web:IWebProfile` (only subtypes of that should be available to the user).
   *
   * @param {String} typeSpec
   * @returns {Promise.<{ defaultType: string, targetType: string }>} promise
   * to be resolved with an object literal containing `BTypeConfig` info
   */
  exports.getTypeConfigInfo = function (typeSpec) {
    return getTypeConfigInfo(String(typeSpec)).then(_.clone); //safe copy
  };

  /**
   * Retrieve information about the given `BTypeConfig.IConfigurable` class.
   * The resolved object will contain the following properties:
   *
   * - `config`: an array of objects corresponding to each key provided by
   *   the `IConfigurable`
   *
   * Each object in the `config` array will contain the following properties:
   *
   * - `key`: the string identifying that config object
   * - `displayName`: a display name string for that key
   * - `facets`: a `baja.Facets` for that config object
   * - `value`: the default `baja.Value` for that config object
   *
   * @param {String} typeConfigSpec
   * @param {String} configurableSpec
   * @returns {Promise} promise to be resolved with an object literal
   * containing `IConfigurable` info
   */
  exports.getConfigurableInfo = function (typeConfigSpec, configurableSpec) {
    // { config: [ { 'displayName', 'facets', 'key', 'value' } ] }
    // since we're memoizing, we need to return a safe copy
    return getConfigurableInfo(String(typeConfigSpec), String(configurableSpec))
      .then(function (obj) {
        return Promise.all(_.map(obj.config, decodeConfigObject));
      })
      .then(function (config) {
        return { config: config };
      });
  };

  function decodeConfigObject(obj) {
    var facets = obj.facets;
    var value = obj.value;
    return Promise.all([ decodeAsync(facets), decodeAsync(value) ])
      .spread(function (facets, value) {
        return _.extend({}, obj, { facets: facets, value: value });
      });
  }

  return (exports);
});
