/**
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

define(['baja!', 'Promise'], function (baja, Promise) {
  'use strict';

////////////////////////////////////////////////////////////////
// Exports
////////////////////////////////////////////////////////////////


  /**
   * RPC calls for working with server-side user data. All data will be
   * automatically keyed to the current user.
   *
   * @private
   * @exports nmodule/webEditors/rc/servlets/userData
   */
  var exports = {};

  /**
   * Retrieve existing data for this key from the station.
   *
   * @param {string} key
   * @param {Object} [params]
   * @param {baja.comm.Batch} [params.batch]
   * @returns {Promise.<string>}
   */
  exports.get = function (key, params) {
    return isInvalidKey(key) || doRpc('get', [ key ], params);
  };

  /**
   * Store a value on the station for this key.
   *
   * @param {string} key
   * @param {string} data
   * @param {Object} [params]
   * @param {baja.comm.Batch} [params.batch]
   * @returns {Promise.<string>}
   */
  exports.put = function (key, data, params) {
    return isInvalidKey(key) || isInvalidData(data) ||
      doRpc('put', [ key, data ], params);
  };

  /**
   * Delete the value on the station for this key.
   *
   * @param {string} key
   * @param {Object} [params]
   * @param {baja.comm.Batch} [params.batch]
   * @returns {Promise}
   */
  exports.del = function (key, params) {
    return isInvalidKey(key) || doRpc('delete', [ key ], params);
  };


////////////////////////////////////////////////////////////////
// Support functions
////////////////////////////////////////////////////////////////

  function doRpc(method, args, params) {
    return baja.rpc({
      ord: 'service:web:UserDataConfig',
      method: method,
      args: args
    })
      .catch(function (err) {
        return null;
      });
  }

  function isInvalidKey(key) {
    if (!key || typeof key !== 'string') {
      return Promise.reject(new Error('key required'));
    }
  }

  function isInvalidData(data) {
    if (typeof data !== 'string') {
      return Promise.reject(new Error('data required'));
    }
  }

  return (exports);
});
