/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/*jshint browser: true */
/*global niagara: false */

require.config({
  shim: {
    'nmodule/js/rc/jquery/contextMenu/jquery.contextMenu': { deps: ['jquery'] },
    'nmodule/js/rc/jquery/contextMenu/jquery.ui.position': { deps: ['jquery'] }
  }
});

/**
 * @private
 * @module nmodule/webEditors/rc/wb/menu/CommandGroupContextMenu
 */
define(['baja!',
        'bajaux/commands/Command',
        'jquery',
        'nmodule/js/rc/jquery/contextMenu/jquery.contextMenu',
        'nmodule/js/rc/jquery/contextMenu/jquery.ui.position',
        'nmodule/js/rc/log/Log',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/fe',
        'css!nmodule/js/rc/jquery/contextMenu/jquery.contextMenu'], function (
         baja,
         Command,
         $,
         $cm,
         $ui,
         Log,
         Promise,
         _,
         fe) {

  'use strict';

  var DATA_KEY = 'contextMenuConfig';

////////////////////////////////////////////////////////////////
// Support functions
////////////////////////////////////////////////////////////////

  //TODO: can we stipulate that an editor add the contextMenu class to anything it wants right clickable? or let editor choose?
  /**
   * Get the CSS selector appropriate for the loaded editor.
   *
   * @inner
   * @param ed
   * @returns {string}
   */
  function getSelector(ed) {
    return ed.getContextMenuSelector ? ed.getContextMenuSelector() : '.contextMenu';
  }

  //TODO: just use pageX/pageY if Workbench ever supports it
  function getPageX(e) {
    if (e.touches && e.touches.length > 0) {
      return e.touches[0].pageX;
    } else {
      return e.pageX || e.clientX + $(window).scrollLeft();
    }
  }

  function getPageY(e) {
    if (e.touches && e.touches.length > 0) {
      return e.touches[0].pageY;
    } else {
      return e.pageY || e.clientY + $(window).scrollTop();
    }
  }


  //TODO: disable a second right click while toItems is working

  /**
   * Get the command from the items object with the specified key.
   *
   * @inner
   * @param {Object} items items object from `toItems()`
   * @param {String} key
   * @returns {module:bajaux/commands/Command}
   */
  function getCommand(items, key) {
    var result = items,
        keys = key.split('|').slice(0, -1);

    _.each(keys, function (key) {
      result = result[key].items;
    });

    return result[key].cmd;
  }

  /**
   * Convert a command to an item object that `jQuery-contextMenu` can handle.
   *
   * @inner
   * @param {module:bajaux/commands/Command} cmd
   * @param {String} key the key string to be assigned to this command
   * @returns {Promise} promise to be resolved with the object literal
   */
  function commandToItem(cmd, key) {
    var div = $('<div/>'),
        iconSpan = $('<span class="icon"/>').appendTo(div),
        displaySpan = $('<span class="display" />').appendTo(div);

    return cmd.toDisplayName()
      .then(function (displayName) {
        displaySpan.text(displayName);
        //TODO: jQuery-contextMenu expects icons to use .icon-iconName CSS format. so should we
        return fe.buildFor({
          value: baja.Icon.make([ cmd.getIcon() ]),
          dom: iconSpan,
          formFactor: 'mini'
        })
          .then(function () {
            return {
              cmd: cmd,
              key: key,
              disabled: !cmd.isEnabled(),
              name: div.html(),
              isHtmlName: true
            };
          });
      });
  }

  /**
   * Convert a child `CommandGroup` to an object literal that
   * `jQuery-contextMenu` can handle, with child objects for child command
   * groups and commands.
   *
   * @inner
   * @param {module:bajaux/commands/CommandGroup} group
   * @param {String} key the string key to be assigned to this group
   * @returns {Promise} promise to be resolved with the object literal
   */
  function groupToItems(group, key) {
    return Promise.all(_.map(group.getChildren(), function (kid, i) {
      if (kid instanceof Command) {
        return commandToItem(kid, (key ? key + '|' : '') + 'cmd' + i);
      }
      return groupToItems(kid, (key ? key + '|' : '') + 'grp' + i);
    }))
      .then(function (items) {
        var itemsObj = {};

        _.each(items, function (item) {
          itemsObj[item.key] = item;
        });

        return group.toDisplayName()
          .then(function (displayName) {
            return {
              name: displayName,
              isHtmlName: true,
              disabled: !items.length,
              items: items.length && itemsObj,
              key: key
            };
          });
      });
  }

  /**
   * Convert the root command group to an object literal to be fed directly to
   * the `$.contextMenu` call.
   *
   * @inner
   * @param {module:bajaux/commands/CommandGroup} group the root group - direct
   * output from `$toContextMenu()`
   * @returns {Promise} promise to be resolve with an object literal
   * to be used as the `items` parameter to `$.contextMenu`
   */
  function rootToItems(group) {
    return groupToItems(group, '').then(_.property('items'));
  }

////////////////////////////////////////////////////////////////
// Exports
////////////////////////////////////////////////////////////////

  /**
   * An add-on that enables right-click menus on editors with the
   * `contextMenuSupport` mixin. The context menu `CommandGroup` will be
   * shown using the `jQuery-contextMenu` library.
   *
   * If the editor does not have the `contextMenuSupport` mixin, you should
   * subclass `CommandGroupContextMenu` and implement `toContextMenu` yourself.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/wb/menu/CommandGroupContextMenu
   * @param {module:nmodule/webEditors/rc/fe/baja/BaseEditor} editor an editor
   * that in most cases should have the `contextMenuSupport` mixin added
   * @see module:nmodule/webEditors/rc/wb/mixin/ContextMenuSupport
   */
  var CommandGroupContextMenu = function CommandGroupContextMenu(editor) {
    if (!editor) {
      throw new Error('Editor required');
    }
    this.$ed = editor;
  };

  /**
   * Set up `jQuery-contextMenu` on this element, but bypass the built-in
   * event handling as we need to do some async work before menu shows up.
   * The `mousedown` handler we arm ourselves will do some async work, sock away
   * the data on the DOM element itself, and then manually trigger the
   * `contextMenu` call. `contextMenu` will then grab that data and use it
   * to build the menu on demand.
   *
   * @private
   * @param jq
   * @param selector
   */
  CommandGroupContextMenu.prototype.$armContextMenuAsync = function (jq, selector) {
    jq.contextMenu({
      className: this.$getClassName(),
      selector: selector,
      trigger: 'none',
      reposition: false,
      build: function ($trigger, e) {
        var data = $trigger.data(DATA_KEY) || false;
        $trigger.removeData(DATA_KEY);
        return data;
      }
    });
  };

  /**
   * Return a class name to apply to the menu root.
   *
   * @private
   * @returns {string}
   */
  CommandGroupContextMenu.prototype.$getClassName = function () {
    return 'hide-root-icons';
  };

  /**
   * Asynchronously resolve the items to show in the context menu, then show it.
   *
   * @private
   * @param {jQuery.Event} e the right click event
   * @returns {boolean} false if the event needs to stop propagation
   */
  CommandGroupContextMenu.prototype.$doContextMenu = function (e) {
    var jq = $(e.currentTarget);
    this.toItems(e).then(function (items) {
      if (!items) { return; }
      jq.data(DATA_KEY, {
        callback: function (key, options) {
          getCommand(items, key).invoke()
            .catch(Log.error);
        },
        items: items
      });

      jq.contextMenu({
        //use 1 because if you wind up with 0, jQuery-contextMenu treats it as falsy
        x: getPageX(e) || 1,
        y: getPageY(e) || 1
      });
    })
      .catch(Log.error);

    return false;
  };


  /**
   * Instantiates an items config object to be passed to the `$.contextMenu`
   * call. Config will be built up using the `CommandGroup` provided through
   * the `contextMenuSupport` mixin.
   *
   * @param {jQuery.Event} e the right-click event
   * @returns {Promise} promise to be resolved with an `items` config
   * object. See the `jQuery-contextMenu` documentation for details.
   */
  CommandGroupContextMenu.prototype.toItems = function (e) {
    return Promise.resolve(this.toContextMenu(e)).then(function (group) {
      return group && rootToItems(group);
    });
  };

  CommandGroupContextMenu.prototype.toContextMenu = function (e) {
    return this.$ed.$toContextMenu(e);
  };

  /**
   * JavaFX doesn't fire `contextmenu` events unless actually showing the Swing
   * context menu, which we disable (see BWebBrowser constructor). So fake it -
   * a right click will trigger `contextmenu` as expected. Only if WB/JavaFX
   * browser is detected.
   *
   * @param {JQuery} jq
   * @param {string} selector
   */
  CommandGroupContextMenu.shimContextMenuForWorkbench = function (jq, selector) {
    if (CommandGroupContextMenu.$isWorkbench() && CommandGroupContextMenu.$isJavaFX()) {
      jq.on('mousedown', selector, function (e) {
        if (e.which === 3) {
          $(e.target).trigger($.Event('contextmenu', {
            which: 3,
            originalEvent: e.originalEvent,
            data: e.data,
            pageX: e.pageX,
            pageY: e.pageY,
            offsetX: e.offsetX,
            offsetY: e.offsetY
          }));
        }
      });
    }
  };

  /**
   * @private
   * @returns {boolean}
   */
  CommandGroupContextMenu.$isWorkbench = function () {
    return typeof niagara !== 'undefined' && niagara.env.type === 'wb';
  };

  /**
   * @private
   * @returns {boolean}
   */
  CommandGroupContextMenu.$isJavaFX = function () {
    return navigator.userAgent.match(/JavaFX/);
  };

  /**
   * Arm event handlers on the loaded editor's DOM element to show the
   * context menu on right-click.
   *
   * @param {Object} [params]
   * @param {string} [params.selector='.contextMenu'] the selector from which to
   * delegate contextmenu events. If not given,
   * `editor.getContextMenuSelector()` will be called, if present; otherwise
   * default selector will be used.
   */
  CommandGroupContextMenu.prototype.arm = function (params) {
    var that = this,
        ed = that.$ed,
        selector = (params && params.selector) || getSelector(ed),
        jq = ed.jq();

    function doContextMenu(e) {
      return that.$doContextMenu(e);
    }

    CommandGroupContextMenu.shimContextMenuForWorkbench(jq, selector);

    if (ed.armContextMenu) {
      ed.armContextMenu(doContextMenu);
    } else {
      jq.on('contextmenu', selector, function (e) {
        if (e.which === 3) { return doContextMenu(e); }
        return false;
      });
    }

    that.$armContextMenuAsync(jq, selector);
  };

  CommandGroupContextMenu.prototype.destroy = function () {

  };

  /**
   * $.contextMenu is kind enough to expose its event handlers to us, so we
   * override layerClick which is the event when we click the transparent
   * fullscreen div shown when a context menu is open.
   *
   * The default implementation of it includes a bunch of
   * repositioning/retriggering logic that works great during normal use when
   * $.contextMenu is used synchronously, but since we have to defer displaying
   * the context menu until after we've had a chance to do network calls for
   * a menu agent, it falls apart for our use case.
   *
   * So we reimplement and simplify it. The only thing it will do now is listen
   * for a click and re-trigger it on the element that the user would have
   * clicked had the transparent div not been in the way.
   */
  $.contextMenu.handle.layerClick = function (e) {
    var root = $(this).data('contextMenuRoot'),
        $win = $(window),
        pageX = getPageX(e),
        pageY = getPageY(e),
        target,
        layerAteMouseup;

    // find the element that would've been clicked, wasn't the layer in the way
    if (document.elementFromPoint) {
      root.$layer.hide();
      target = document.elementFromPoint(pageX - $win.scrollLeft(), pageY - $win.scrollTop());
      root.$layer.show();
    }

    function clickName(which) { return which === 3 ? 'contextmenu' : 'click'; }

    function retriggerOnTarget(name) {
      $(target).trigger($.Event(name, {
        which: e.which,
        pageX: pageX,
        pageY: pageY,
        shiftKey: e.shiftKey,
        ctrlKey: e.ctrlKey,
        originalEvent: e.originalEvent
      }));
    }

    var targetListener = function (e) {
      if (!layerAteMouseup) { retriggerOnTarget(clickName(e.which)); }
    };

    root.$layer.one('mouseup', function (e) { layerAteMouseup = e.which || 1; });
    $(target).one('mouseup', targetListener);

    e.preventDefault();
    e.stopImmediatePropagation();

    if (target) {
      root.$trigger.one('contextmenu:hidden', function () {
        retriggerOnTarget('mousedown');
        if (layerAteMouseup) {
          $(target).off('mouseup', targetListener);
          retriggerOnTarget('mouseup');
          retriggerOnTarget(clickName(layerAteMouseup));
        }
      });
    }

    root.$menu.trigger('contextmenu:hide');
  };

  return (CommandGroupContextMenu);
});
