/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/mgr/BatchComponentEditor
 */
define(['baja!',
        'log!nmodule.webEditors.rc.wb.mgr.BatchComponentEditor',
        'bajaux/events',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/fe/fe',
        'nmodule/webEditors/rc/fe/baja/BaseEditor',
        'nmodule/webEditors/rc/fe/baja/util/ComplexDiff',
        'nmodule/webEditors/rc/util/ListSelection',
        'nmodule/webEditors/rc/util/Switchboard',
        'nmodule/webEditors/rc/wb/mgr/MgrSheet',
        'nmodule/webEditors/rc/wb/mgr/model/MgrColumn',
        'nmodule/webEditors/rc/wb/mgr/model/MgrModelCompositeBuilder',
        'nmodule/webEditors/rc/wb/mgr/model/columns/NameMgrColumn',
        'nmodule/webEditors/rc/wb/mgr/model/columns/TypeMgrColumn',
        'nmodule/webEditors/rc/wb/table/Table'], function (
         baja,
         log,
         events,
         Promise,
         _,
         fe,
         BaseEditor,
         ComplexDiff,
         ListSelection,
         Switchboard,
         MgrSheet,
         MgrColumn,
         MgrModelCompositeBuilder,
         NameMgrColumn,
         TypeMgrColumn,
         Table) {

  'use strict';

  var MODIFY_EVENT = events.MODIFY_EVENT,
      logError = log.severe.bind(log);

  function newCopy(value, row, column) {
    if (value instanceof ComplexDiff) {
      var copy = column.getValueFor(row),
          str = copy.getDisplay();
      copy = copy.newCopy();
      copy.getDisplay = function () { return str; };
      return value.apply(copy);
    }
    return typeof value.newCopy === 'function' ? value.newCopy() : value;
  }

  function proposeNewCopy(ed, value, row, column) {
    return Promise.resolve(newCopy(value, row, column))
      .then(function (copy) {
        return column.propose(copy, row);
      })
      .then(function () {
        if (column instanceof NameMgrColumn) {
          NameMgrColumn.setHasUserDefinedValue(row, true);
        } else if (column instanceof TypeMgrColumn) {
          // If the row's type has changed, force an update of the property sheet
          return ed.$onSelectionChanged();
        }
      });
  }

  /**
   * Editor intended for use with a `MgrModel`, allowing multiple components
   * to be edited simultaneously.
   *
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/wb/mgr/BatchComponentEditor
   * @extends module:nmodule/webEditors/rc/fe/baja/BaseEditor
   */
  var BatchComponentEditor = function BatchComponentEditor() {
    var that = this,
        selection = that.$selection = new ListSelection();

    BaseEditor.apply(that, arguments);

    selection.select(0);

    selection.on('changed', function () {
      that.$onSelectionChanged().catch(logError);
    });

    new Switchboard(that)
      .allow('$onSelectionChanged')
      .oneAtATime().onRepeat().preempt();
  };
  BatchComponentEditor.prototype = Object.create(BaseEditor.prototype);
  BatchComponentEditor.prototype.constructor = BatchComponentEditor;

  /**
   * Update the property sheet with values coalesced from the selected rows.
   *
   * @private
   * @returns {Promise}
   */
  BatchComponentEditor.prototype.$onSelectionChanged = function () {
    return this.$getPropertySheet().getBuilder().buildAll();
  };

  /**
   * Get the element in which the component table is shown.
   *
   * @private
   * @returns {jQuery}
   */
  BatchComponentEditor.prototype.$getTableElement = function () {
    return this.jq().children('.editTable').children('table');
  };

  /**
   * Get the element in which the property sheet is shown.
   *
   * @private
   * @returns {jQuery}
   */
  BatchComponentEditor.prototype.$getPropertySheetElement = function () {
    return this.jq().children('.editSheet').children('div');
  };

  /**
   * Get the component table showing all the components to be edited.
   *
   * @private
   * @returns {module:nmodule/webEditors/rc/wb/table/Table}
   */
  BatchComponentEditor.prototype.$getTable = function () {
    return this.$getTableElement().data('widget');
  };

  /**
   * Get the property sheet allowing individual values to be edited.
   *
   * @private
   * @returns {module:nmodule/webEditors/rc/wb/PropertySheet}
   */
  BatchComponentEditor.prototype.$getPropertySheet = function () {
    return this.$getPropertySheetElement().data('widget');
  };

  /**
   * Take a field from the property sheet and apply the entered
   * values as proposed - not committed - changes to the table model.
   *
   * @private
   * @param {String} key the editor key
   * @returns {Promise} promise to be resolved when entered changes have
   * been applied as proposed changes to the table model
   */
  BatchComponentEditor.prototype.$applySheetToModel = function (key) {
    var that = this,
        sheet = that.$getPropertySheet(),
        builder = sheet.getBuilder(),
        kid = builder.getEditorFor(key),
        selectedRows = builder.$getSelectedRows(),
        mgrModel = that.value();

    if (!kid) {
      return Promise.resolve();
    }

    return kid.validate()
      .then(function (value) {
        var column = mgrModel.getColumn(key),
            isArray = _.isArray(value);

        return Promise.all(selectedRows.map(function (row, i) {
          return proposeNewCopy(that, isArray ? value[i] : value, row, column);
        }));
      })
      .then(function () {
        mgrModel.emit('rowsChanged', selectedRows);
      });
  };


  /**
   * Set up elements to show a component table and property sheet.
   *
   * Arm handlers to pick up on edits to the property sheet, and apply those
   * edits as proposed changes to the table model.
   *
   * Add `BatchComponentEditor` class.
   *
   * @param {JQuery} dom
   */
  BatchComponentEditor.prototype.doInitialize = function (dom) {
    var that = this;

    dom
      .addClass('BatchComponentEditor')
      .html(
        '<div class="editTable"><table class="ux-table no-stripe"></table></div>' +
        '<div class="editSheet"><div></div></div>'
      );

    dom.on(MODIFY_EVENT, '.editSheet', function (e, ed, modifiedEd) {
      that.$applySheetToModel(modifiedEd.jq().data('key')).catch(logError);
    });
  };

  /**
   * Load the component table and property sheet.
   *
   * @param {module:nmodule/webEditors/rc/wb/mgr/model/MgrModel} model
   * @returns {Promise}
   */
  BatchComponentEditor.prototype.doLoad = function (model) {
    var that = this,
        selection = that.$selection;
    return that.getOrdBase()
      .then(function (ordBase) {
        return Promise.join(
          fe.buildFor({
            selection: selection,
            dom: that.$getTableElement(),
            properties: { hideUnseenColumns: false },
            value: model,
            type: Table
          }),
          fe.buildFor({
            data: {
              builder: new MgrModelCompositeBuilder({
                selection: selection,
                model: model
              })
            },
            dom: that.$getPropertySheetElement(),
            value: model,
            properties: {
              allowHyperlink: false,
              ordBase: ordBase,
              showHeader: false,
              showControls: false,
              nested: true //NCCB-23224: Prevent PropertySheetSave on Enter Key
            },
            type: MgrSheet
          })
        );
      });
  };

  function batchCommit(column, data, row, ed, batch) {
    var readyToCommit,
        params = { editor: ed, batch: batch };

    readyToCommit = new Promise(function (resolve /*, reject*/) {
      params.progressCallback = function (msg) {
        if (msg === MgrColumn.COMMIT_READY) {
          resolve();
        }
      };
    });

    return [ readyToCommit, column.commit(data, row, params) ];
  }

  /**
   * Commit any proposed changes back to the underlying `MgrModel`.
   *
   * @returns {Promise}
   */
  BatchComponentEditor.prototype.doSave = function () {
    var that = this,
        sheet = that.$getPropertySheet(),
        builder = sheet.getBuilder(),
        model = this.value(),
        rows = model.getRows(),
        columns = model.getColumns();

    return Promise.resolve(builder.getKeys())
      .then(function (keys) {
        return Promise.all([keys, Promise.all(keys.map(function (key) {
          var ed = builder.getEditorFor(key);
          return Promise.resolve(ed.shouldValidate() && ed.validate())
            .then(function () {
              var column = model.getColumn(key),
                data = _.map(rows, function (row) {
                  return row.data('proposed.' + key) || null;
                });

              return column.mgrValidate(model, data, { editor: ed });
            });
        })) ]);
      })
      .then(function () {
        // we must batch together the column commits, waiting for the
        // COMMIT_READY callback for each one before we commit the batch.
        // this logic is almost identical to bajaux/mixin/batchSaveMixin.
        // With the addition of manager discovery support, this iteration
        // is now done by column name, rather than by keys from the builder,
        // as we want to commit rows with proposed values for non-editable
        // columns: these may have had their proposed values set from
        // adding a discovered object.

        var batch = new baja.comm.Batch(),
          results = _.map(columns, function (column) { //need this to run sync
            var key = column.getName();
            return _.compact(_.map(rows, function (row) {
              var data = row.data('proposed.' + key);
              if (data !== undefined) {
                var ed = builder.getEditorFor(key);
                return batchCommit(column, data, row, ed, batch);
              }
            }));
          });

        results = _.flatten(results, /*shallow*/ true);

        var savePromises = _.map(results, function (arr) { return arr[1]; }),
            commitPromises = _.map(results, function (arr) { return arr[0]; });

        Promise.all(commitPromises)
          .then(function () {
            batch.commit();
          });

        return Promise.all(savePromises);
      });
  };

  /**
   * Remove `BatchComponentEditor` class, destroy table and property sheet.
   * @returns {Promise}
   */
  BatchComponentEditor.prototype.doDestroy = function () {
    var table = this.$getTable(),
        sheet = this.$getPropertySheet();

    this.jq().removeClass('BatchComponentEditor');
    return Promise.join(table && table.destroy(), sheet && sheet.destroy());
  };

  BatchComponentEditor.prototype.doLayout = function () {
    return this.getChildWidgets().layoutAll();
  };

  return (BatchComponentEditor);
});
