/*
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 */

/* jshint browser: true */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/mgr/commands/AddCommand
 */
define(['baja!',
        'lex!webEditors',
        'Promise',
        'underscore',
        'bajaux/commands/Command',
        'nmodule/webEditors/rc/fe/feDialogs',
        'nmodule/webEditors/rc/wb/mgr/BatchComponentEditor',
        'nmodule/webEditors/rc/wb/mgr/mgrUtils',
        'nmodule/webEditors/rc/wb/mgr/model/MgrModel',
        'nmodule/webEditors/rc/wb/mgr/model/columns/TypeMgrColumn'], function (
        baja,
        lexs,
        Promise,
        _,
        Command,
        feDialogs,
        BatchComponentEditor,
        mgrUtils,
        MgrModel,
        TypeMgrColumn) {

  'use strict';

  var webEditorsLex = lexs[0],
      getLearnTableSelectedSubjects = mgrUtils.getLearnTableSelectedSubjects,
      setProposedValues = mgrUtils.setProposedDiscoveryValues,
      toMgrTypeInfos = mgrUtils.toMgrTypeInfos;

  function getComponentNames(comps) {
    return _.invoke(comps, 'getName');
  }

  function getRowSubjects(rows) {
    return _.invoke(rows, 'getSubject');
  }

  function removeRowSubjectsFromContainer(container, model) {
    _.each(getRowSubjects(model.getRows()), function (comp) {
      container.remove(comp);
    });
  }

  /**
   * From the available types stored against the row, find the MgrTypeInfo
   * that corresponds to the type of the row's subject.
   */
  function getTypeInfoForRow(row) {
    var comp = row.getSubject(),
        type = comp.getType(),
        typeInfos = TypeMgrColumn.getAvailableTypes(row);

    return _.find(typeInfos, function (t) {
      return type.equals(t.getType());
    });
  }

  /**
   * Creates an array of slot names for the given rows, based on the type of
   * the row's subject component.
   */
  function generateSlotNamesForRows(rows, origContainer) {
    var slotNames = [];

    _.each(rows, function (row) {
      var typeInfo = getTypeInfoForRow(row),
          uniqueName = origContainer.getUniqueName(typeInfo.toSlotName()),
          slotName,
          slotNum = uniqueName.match(/\d*$/)[0] || 0;

      uniqueName = uniqueName.replace(/\d*$/, '');

      while ((origContainer.has(slotName = uniqueName + (slotNum || ''))) || (_.contains(slotNames, slotName))) {
        slotNum++;
      }

      slotNames.push(slotName);
    });

    return slotNames;
  }

  /**
   * Returns a new `MgrModel` containing rows representing the items
   * to be added from the discovery table to the database table. This will
   * be used to present the batch component editor before the rows
   * are added to the main model's component source.
   */
  function makeModelForNewRows(mgr, rows) {
    var mgrModel = mgr.getModel(),
        source = baja.$('baja:Folder'),
        origContainer = mgr.value(),
        slotNames = generateSlotNamesForRows(rows, origContainer),
        addModel = new MgrModel({
          componentSource: source,
          columns: mgrModel.getColumns()
        });

     _.each(rows, function (row, index) {
      source.add({
        slot: slotNames[index],
        value: row.getSubject()
      });
    });

    return addModel.insertRows(rows, 0)
      .then(_.constant(addModel));
  }

  /**
   * Shows the `BatchComponentEditor` in a dialog with the `Rows`
   * created from the discovery nodes dragged or selected in the
   * learn table. Once the edit is confirmed, the newly dropped
   * components are added to the station.
   *
   * @param {module:nmodule/webEditors/rc/wb/mgr/Manager} mgr the manager instance
   * @param {Array.<module:nmodule/webEditors/rc/wb/table/model/Row>} rows
   */
  function addNewInstances(mgr, rows) {
    var dbModel = mgr.getModel(),
        baseComponent;

    return Promise.all([ mgr.getOrdBase(), makeModelForNewRows(mgr, rows) ])
      .spread(function (ordBase, editorModel) {
        baseComponent = ordBase;
        return feDialogs.showFor({
          value: editorModel,
          properties: { ordBase: baseComponent },
          type: BatchComponentEditor
        })
        .then(function (addModel) {
          var newRows, newInstances, existingInstances, duplicate;
          if (addModel) {

            // Look at the names of the components we are being asked to add.
            // If any of the names match existing component names, we'll go
            // back and present the add dialog again, to let the user change
            // the name to a non-clashing one.

            newRows = addModel.getRows();
            newInstances = getRowSubjects(newRows);

            existingInstances = getRowSubjects(dbModel.getRows());
            duplicate = findDuplicateSlotName(existingInstances, newInstances);

            if (!duplicate) {
              return dbModel.addInstances(newInstances);
            } else {
              return feDialogs.error(webEditorsLex.get('mgr.duplicateName', duplicate))
                .then(function () {
                  removeRowSubjectsFromContainer(mgr.value(), addModel);
                  return addNewInstances(mgr, addModel.getRows());
                });
            }
          }
        })
        .catch(feDialogs.error);
      });
  }

  /**
   * Search through the two arrays of existing components and new components
   * to be added, looking for any duplicates between the name names and the
   * existing ones. If we find a clash, return the name of the first duplicate
   * so we can show it in an error message.
   *
   * @param {Array.<baja.Component>} existingComps
   * @param {Array.<baja.Component>} newComps
   * @returns {String} the first duplicate name, or null if there were no duplicates found.
   */
  function findDuplicateSlotName(existingComps, newComps) {
    var existingNames = getComponentNames(existingComps),
        newNames = getComponentNames(newComps),
        duplicates = _.intersection(existingNames, newNames);

    return duplicates.length ? _.first(duplicates) : null;
  }

  /**
   * Command for adding discovered items to the station database. The command action can be
   * invoked directly by clicking on the button, or can be invoked by a drag and drop from
   * the discovery table to the main table.
   *
   * @private
   * @class
   * @extends module:bajaux/commands/Command
   * @alias module:nmodule/webEditors/rc/wb/mgr/commands/AddCommand
   * @param {module:nmodule/webEditors/rc/wb/mgr/MgrLearn} mgr
   */
  var AddCommand = function AddCommand(mgr) {

    Command.call(this, {
      module: 'webEditors',
      lex: 'commands.mgr.add',
      enabled: false,

      /**
       * When this command is invoked, we should take the selected components in the discovery
       * table and present a dialog for editing them, before adding them to the station.
       *
       * @alias module:nmodule/webEditors/rc/wb/mgr/commands/AddCommand#invoke
       * @returns {Promise}
       */
      func: function () {
        var selection,
            dbModel = mgr.getModel();

        // Obtain the current selection from the discovery table - these
        // are the items that were dragged or selected. We then map those items to
        // new Rows for the main manager table, and show the editor dialog
        // to configure them before they are added to the station.

        selection = getLearnTableSelectedSubjects(mgr);

        if (!selection.length) {
          return Promise.resolve();
        }

        // First we query the manager for the types for each item
        // being dropped. If the value resolved from the call to
        // getTypesForDiscoverySubject() is not a MgrTypeInfo (or an
        // array of), we ensure we create them ourselves. We then
        // create a new instance, which unless the concrete manager
        // has overridden the function, will simply call the
        // #newInstance() function on the first MgrTypeInfo we have.

        return Promise.all(selection.map(function (discovery) {
          var row, typeInfos;

          return Promise.resolve(mgr.getTypesForDiscoverySubject(discovery))
            .then(toMgrTypeInfos)
            .then(function (ts) {
              typeInfos = ts;
              return mgr.newInstanceFromDiscoverySubject(discovery, typeInfos);
            })
            .then(function (comp) {
              row = dbModel.makeRow(comp);

              TypeMgrColumn.setAvailableTypes(row, typeInfos);
              TypeMgrColumn.setDiscoveryData(row, discovery);

              return setProposedValues(mgr, dbModel, discovery, row);
            })
            .then(function () {
              return row;
            });
        }))
        .then(function (rows) {
          return addNewInstances(mgr, rows);
        });
      }
    });
  };

  AddCommand.prototype = Object.create(Command.prototype);
  AddCommand.prototype.constructor = AddCommand;

  return (AddCommand);
});
