/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/*jshint browser: true */

/**
 * @module nmodule/webEditors/rc/wb/table/Table
 */
define(['log!nmodule.webEditors.rc.wb.table',
        'jquery',
        'Promise',
        'underscore',
        'nmodule/js/rc/switchboard/switchboard',
        'nmodule/webEditors/rc/fe/BaseWidget',
        'nmodule/webEditors/rc/util/htmlUtils',
        'nmodule/webEditors/rc/util/ListSelection',
        'nmodule/webEditors/rc/wb/table/model/TableModel',
        'nmodule/js/rc/log/Log',
        'nmodule/webEditors/rc/wb/table/pagination/PaginationWidget'], function (
        tableLog,
        $,
        Promise,
        _,
        switchboard,
        BaseWidget,
        htmlUtils,
        ListSelection,
        TableModel,
        Log,
        PaginationWidget) {

  'use strict';

  var preventSelectOnShiftClick = htmlUtils.preventSelectOnShiftClick;

  var logError = tableLog.severe.bind(tableLog);

  var TABLE_HTML = '<thead class="ux-table-head"></thead>' +
      '<tbody></tbody>' +
      '<tfoot class="ux-table-foot"></tfoot>',
    QUEUE_UP = { allow: 'oneAtATime', onRepeat: 'queue' };

////////////////////////////////////////////////////////////////
// Support functions
////////////////////////////////////////////////////////////////

  function toCssClass(column) {
    return 'js-col-' + column.getName().replace(' ', '_');
  }

  function toCell(table, column, row) {
    var dom = $('<td/>')
      .addClass(toCssClass(column));

    if (table.$isHideUnseen() && column.isUnseen()) {
      dom.hide();
    }

    return Promise.resolve(table.buildCell(column, row, dom))
      .then(function () {
        return dom;
      });
  }

  function toTableRow(table, row, columns, selected) {
    return Promise.all(columns.map(function (column) {
      return toCell(table, column, row);
    }))
      .then(function (tds) {
        return $('<tr class="ux-table-row"></tr>')
          .toggleClass('selected', selected)
          .data('columns', columns.slice())
          .data('row', row)
          .html(tds);
      })
      .then(function (dom) {
        return table.finishBuildingRow(row, dom);
      });
  }

  function destroyCells(table, tbody) {
    var columns = tbody.data('columns');
    tbody.removeData('columns');

    function mapDom(elem, fun) {
      return elem.map(fun).get();
    }

    return Promise.all(mapDom(tbody.find("tr"), function () {
      var $tr = $(this),
          row = $tr.data('row');

      $tr.removeData('row');
      return mapDom($tr.find("td"), function () {
        var $td = $(this),
            col = columns[this.cellIndex];

        return Promise.resolve(table.destroyCell(col, row, $td));
      });
    }))
    .then(function () {
      tbody.html('');
    });
  }

  function updateTbody(table, tableModel) {
    //TODO: be smart about inserts/removes

    var tbody = table.$getTbody(tableModel),
      selection = table.$getSelection(),
      rows = tableModel.getRows(),
      columns = tableModel.getColumns();

    return destroyCells(table, tbody)
      .then(function () {
        return Promise.all(rows.map(function (row, i) {
          return toTableRow(table, row, columns, selection.isSelected(i));
        }));
      })
      .then(function (rows) {
        tbody.data('columns', columns.slice()).html(rows);
      });
  }

  function addRows(table, tableModel, addedRows, index) {
    var tbody = table.$getTbody(tableModel),
      columns = tableModel.getColumns(),
      selection = table.$getSelection();

    return Promise.all(_.map(addedRows, function (row) {
      return toTableRow(table, row, columns, false);
    }))
      .then(function (trs) {
        if (index === 0) {
          tbody.prepend(trs);
        } else {
          tbody.children(':eq(' + (index - 1) + ')').after(trs);
        }
        selection.insert(index, addedRows.length);
      });
  }

  function removeIt(elem) {
    elem.remove();
  }

  function removeRows(table, tableModel, removedRows, indices) {
    var tbody = table.$getTbody(tableModel),
      selection = table.$getSelection(),
      kids = tbody.children(),
      trs = $(_.map(indices, function (i) {
        return $(kids[i]);
      }));

    return table.getChildWidgets(trs).destroyAll()
      .then(function () {
        _.each(trs, removeIt);
        selection.remove(indices);
      });
  }

  function updateThead(table, tableModel) {
    var thead = table.$getThead(),
      isFixedHeaders = table.$isFixedHeaders(),
      isHideUnseen = table.$isHideUnseen(),
      columns = tableModel.getColumns();

    return Promise.all(columns.map(function (column) {
      return Promise.resolve(column.toDisplayName())
        .then(function (displayName) {
          var th = $('<th/>')
            .text(displayName)
            .data('column', column)
            .toggle(!isHideUnseen || !column.isUnseen())
            .toggleClass('sortable', column.isSortable())
            .addClass(toCssClass(column));
          if (isFixedHeaders) {
            $('<div class="headerDuplicate ux-table-head-cell" aria-hidden="true"></div>')
              .text(displayName)
              .appendTo(th);
          }
          return th;
        });
    }))
      .then(function (ths) {
        thead.html(ths);
      });
  }

  //TODO: MgrColumn#toSortKey equivalent
  function sortByDataKey(model, dataKey, desc) {
    var lessThan = desc ? 1 : -1,
      greaterThan = desc ? -1 : 1;

    model.sort(function (row1, row2) {
      var display1 = row1.data(dataKey),
        display2 = row2.data(dataKey);
      return display1 === display2 ? 0 :
        display1 < display2 ? lessThan : greaterThan;
    });
  }

  /**
   * This is ugly but I am convinced it is impossible to do with pure CSS.
   *
   * When scrolling with a fixed header, in most browsers a margin-left on the
   * header dups to counteract the container's scrollLeft does the trick. But
   * in Firefox it somehow doubles up on the margin so the headers go sailing
   * off the left side of the table. But you can't just halve the margin - it's
   * correctly positioned without the margin. But you ALSO can't just set the
   * margin to 0 because then it doesn't scroll! So the solution that Firefox
   * likes is apply the margin, force a reflow, then remove the margin, and
   * Firefox then snaps the header back to its correct position of its own
   * accord.
   *
   * TODO: reproduction steps and Firefox bug.
   *
   * @param {module:nmodule/webEditors/rc/wb/table/Table} table
   * @param {JQuery} dom
   */
  function applyFixedHeaderScrolling(table, dom) {
    // i tried to do this without browser sniffing by detecting when the headers
    // got out of whack. swear i did.
    var isFirefox = navigator.userAgent.toLowerCase().match('firefox');
    var onScroll = function onScroll() {
      var headerDups = table.$getThead().find('.headerDuplicate'),
        scrollLeft = this.scrollLeft;
      headerDups.css('marginLeft', -scrollLeft);

      if (isFirefox) {
        headerDups.offset(); //force reflow
        headerDups.css('marginLeft', '0.1px');
      }
    };

    dom.children('.tableContainer').on('scroll', onScroll);
  }


////////////////////////////////////////////////////////////////
// Table
////////////////////////////////////////////////////////////////

  /**
   * API Status: **Development**
   *
   * Table widget.
   *
   * It supports the following `bajaux` `Properties`:
   *
   * - `fixedHeaders`: (boolean) set to true to allow scrolling the table body
   *   up and down while the headers remain fixed. This will only make sense
   *   when the table widget is instantiated in a block-level element, like a
   *   `div`, whose dimensions are constrained.
   * - `hideUnseenColumns`: (boolean) set to `false` to cause columns with the
   *   `UNSEEN` flag to always be shown. Defaults to `true` (unseen columns are
   *   hidden by default).
   *
   * @class
   * @alias module:nmodule/webEditors/rc/wb/table/Table
   * @extends module:nmodule/webEditors/rc/fe/BaseWidget
   * @param {Object} params
   * @param {module:nmodule/webEditors/rc/util/ListSelection} [params.selection] the `ListSelection`
   * to manage which rows are currently selected - if not given, a new one will be constructed.
   * @param {Object} [params.pagination] pagination params
   * @param {module:nmodule/webEditors/rc/wb/table/pagination/PaginationHandler} [params.pagination.paginationHandler]
   * specify a pagination handler instance
   * @param {String} [params.pagination.elSelector] specifies the jQuery selector in which to initialize
   * the pagination widget
   * @param {jQuery} [params.pagination.dom] specifies the jQuery DOM element in which to initialize the
   * pagination widget. Note: this parameter only applies if no element selector has been specified
   * using the 'elSelector' parameter
   */
  var Table = function Table(params) {
    var that = this;

    BaseWidget.apply(that, arguments);

    this.$selection = (params && params.selection) || new ListSelection();

    this.$pagination = (params && params.pagination) || null;

    this.$paginationHandler = (this.$pagination && this.$pagination.paginationHandler) || null;

    this.$paginationWidget = this.$paginationHandler !== null ? new PaginationWidget() : null;

    // if multiple rowsChanged events come in for the same row rapid-fire, ensure
    // that they queue correctly and execute in sequence.
    switchboard(this, {
      //TODO: wrong calls to _.extend
      '$handleRowEvent': _.extend(QUEUE_UP, { notWhile: '$handleColumnEvent' }),
      '$handleColumnEvent': _.extend(QUEUE_UP, { notWhile: '$handleRowEvent' })
    });
  };
  Table.prototype = Object.create(BaseWidget.prototype);
  Table.prototype.constructor = Table;

  Table.prototype.$isFixedHeaders = function () {
    var tagName = this.jq().prop('tagName').toLowerCase();
    var isFixedHeaders = this.properties().getValue('fixedHeaders');
    if (isFixedHeaders && (tagName === 'table' || tagName === 'tbody')) {
      throw new Error("Can only initialize inside a div if fixed headers is true.");
    }
    return tagName !== 'table' && tagName !== 'tbody' &&
      isFixedHeaders !== false;
  };

  /**
   * Get the `ListSelection` representing the currently selected table rows.
   *
   * @private
   * @returns {module:nmodule/webEditors/rc/util/ListSelection}
   */
  Table.prototype.$getSelection = function () {
    return this.$selection;
  };

  Table.prototype.$getPaginationHandler = function () {
    return this.$paginationHandler;
  };

  Table.prototype.$getPaginationWidget = function () {
    return this.$paginationWidget;
  };

  /**
   * Return true if columns with the `UNSEEN` flag should be hidden.
   *
   * @private
   * @returns {boolean}
   */
  Table.prototype.$isHideUnseen = function () {
    return this.properties().getValue('hideUnseenColumns') !== false;
  };

  //noinspection JSUnusedLocalSymbols
  /**
   * Get the table body element(s).
   *
   * @private
   * @param {module:nmodule/webEditors/rc/wb/table/model/TableModel} [tableModel]
   * a table could contain multiple `tbody` tags, one per `TableModel`. If
   * given, this should return a jQuery object of length 1 corresponding only
   * to that `TableModel`. Otherwise, this could return multiple `tbody`
   * elements. By default, there will only be one.
   * @returns {jQuery}
   */
  Table.prototype.$getTbody = function (tableModel) {
    var dom = this.jq();
    switch (dom.prop('tagName').toLowerCase()) {
      case 'table':
        return dom.children('tbody');
      case 'tbody':
        return dom;
      default:
        return dom.find('tbody').eq(0);
    }
  };

  /**
   * Get the table head element.
   *
   * @private
   * @returns {jQuery}
   */
  Table.prototype.$getThead = function () {
    var dom = this.jq();
    switch (dom.prop('tagName').toLowerCase()) {
      case 'table':
        return dom.children('thead');
      case 'tbody':
        return $();
      default:
        return dom.children('.tableContainer').children('table').children('thead');
    }
  };


  Table.prototype.$rebuildTrs = function (tableModel, rows) {
    var that = this,
        selection = that.$getSelection();

    return Promise.all(_.map(rows, function (row) {
      var i = tableModel.getRowIndex(row),
          tr = that.$getTbody(tableModel).children('tr').eq(i),
          selected = selection.isSelected(i);

      return toTableRow(that, row, tableModel.getColumns(), selected)
        .then(function (newTr) {
          tr.replaceWith(newTr);
        });
    }));
  };

  /**
   * Sort table rows based on the display string value provided by the given
   * column.
   *
   * @private
   * @param {module:nmodule/webEditors/rc/wb/table/model/Column} column
   * @param {boolean} desc
   * @returns {Promise}
   */
  Table.prototype.$sortByColumnDisplay = function (column, desc) {
    var that = this,
      dataKey = 'displayString.' + column.getName(),
      model = that.value(),
      rows = model.getRows();

    return Promise.all(rows.map(function (row) {
      var dom = $('<div/>');
      return Promise.resolve(that.buildCell(column, row, dom))
        .then(function () {
          row.data(dataKey, dom.text());
        });
    }))
      .then(function () {
        sortByDataKey(model, dataKey, desc);
        return null; //squelch "promise not returned" warnings from row event handlers
      });
  };

  Table.prototype.$handleRowEvent = function (tableModel, rows, eventName, args) {
    switch (eventName) {
      case 'rowsAdded':
        return addRows(this, tableModel, rows, /* index */ args[0]);
      case 'rowsRemoved':
        return removeRows(this, tableModel, rows, /* indices */ args[0]);
      case 'rowsReordered':
        return updateTbody(this, tableModel);
      case 'rowsChanged':
        return this.$rebuildTrs(tableModel, rows);
    }
  };

  Table.prototype.$handleColumnEvent = function (tableModel, columns, eventName) {
    switch (eventName) {
      case 'columnsAdded':
      case 'columnsRemoved':
        return Promise.join(
          updateThead(this, tableModel),
          updateTbody(this, tableModel)
        );
      case 'columnsFlagsChanged':
        if (this.$isHideUnseen()) {
          var toUpdate = this.$getTbody(tableModel).add(this.$getThead());
          _.each(columns, function (c) {
            toUpdate.find('.' + toCssClass(c)).toggle(!c.isUnseen());
          });
        }
        return;
    }
  };

  /**
   * Initialize the HTML table, creating `thead`, `tbody`, and `tfoot` elements.
   *
   * @param {JQuery} dom
   * @param {Object} params optional initialization parameters
   */
  Table.prototype.doInitialize = function (dom, params) {
    var that = this;

    params = params || {};

    dom.addClass('TableWidget')
      .toggleClass('fixedHeaders', that.$isFixedHeaders());

    preventSelectOnShiftClick(dom);

    var selection = that.$getSelection();

    // Allow users of the table to hook into the default cell building/destruction
    // process for a table, via function parameters that can be specified in fe.buildFor().
    // This is intended to allow clients such as the Manager view to have a bit more
    // control over how the dom content is generated.

    if (params.buildCell) { this.$buildCell = params.buildCell; }
    if (params.destroyCell) { this.$destroyCell = params.destroyCell; }
    if (params.finishBuildingRow) { this.$finishBuildingRow = params.finishBuildingRow; }

    selection.on('changed', function () {
      that.$getTbody().children('tr').each(function (i) {
        $(this).toggleClass('selected', selection.isSelected(i));
      });
    });

    function armHeaderHandlers() {
      dom.on('click', 'thead > th', function () {
        var th = $(this).closest('th'),
            desc = th.data('desc'),
            column = th.data('column');

        if (!(column.isSortable())) { return; }

        that.$getThead().find('th').removeClass('asc desc');
        th.addClass(desc ? 'desc' : 'asc');

        that.$sortByColumnDisplay(column, desc)
          .then(function () {
            th.data('desc', !desc);
          })
          .catch(logError);
      });
    }

    function armRowHandlers(selector) {
      //this must be click and not mousedown, because if it were mousedown then
      //starting a drag would wreck your current selection.
      dom.on('click', selector, function (e) {
        if ($(e.target).is('button')) {
          return false;
        }
        selection.defaultHandler.apply(this, arguments);
      });

      dom.on('dragstart', selector, function (e) {
        var i = $(e.currentTarget).index();
        if (!selection.isSelected(i)) {
          selection.select(i);
        }
      });
    }

    switch (dom.prop('tagName').toLowerCase()) {
      case 'table':
        dom.html(TABLE_HTML);
        armHeaderHandlers();
        armRowHandlers('tbody > tr');
        break;
      case 'tbody':
        armRowHandlers('tr');
        break;
      default:
        dom.html('<div class="tableContainer">' +
          '<table class="ux-table">' +
            TABLE_HTML +
          '</table>' +
          '</div>');

        applyFixedHeaderScrolling(that, dom);

        dom.on('mousedown', '.tableContainer', function (e) {
          var table = dom.children('.tableContainer').children('table')[0];
          if (!$.contains(table, e.target)) {
            selection.clear();
          }
        });

        armHeaderHandlers();
        armRowHandlers('tbody > tr');
    }

    //Initialize the pagination widget
    if (that.$paginationWidget !== null) {
      var paginationDom = that.jq().find(this.$pagination.elSelector);
      if (!paginationDom.length && this.$pagination.dom) {
        paginationDom = this.$pagination.dom;
      }
      return Promise.resolve(paginationDom.length && that.$paginationWidget.initialize(paginationDom));
    }
  };

  /**
   * Get the currently loaded `TableModel`.
   * @returns {module:nmodule/webEditors/rc/wb/table/model/TableModel}
   * @since Niagara 4.6
   */
  Table.prototype.getModel = function () {
    return this.value();
  };

  /**
   * Load in a `TableModel`, immediately rendering all columns and rows. Event
   * handlers will be registered to listen for updates to the table model.
   *
   * @param {module:nmodule/webEditors/rc/wb/table/model/TableModel} tableModel
   * @returns {Promise}
   * @throws {Error} if no TableModel provided
   */
  Table.prototype.doLoad = function (tableModel) {
    if (!(tableModel instanceof TableModel)) {
      throw new Error('TableModel required');
    }

    var that = this;

    function handleRowsEvent(rows, eventName, args) {
      that.$handleRowEvent(tableModel, rows, eventName, args);
    }

    function handleColumnsEvent(columns, eventName) {
      that.$handleColumnEvent(tableModel, columns, eventName);
    }

    //Remove all existing model listeners
    that.$disarmModel();
    //TODO: make these instance methods and protect w/ switchboard
    var handlers = that.$handlers = {};
    _.each({
      rowsAdded: handleRowsEvent,
      rowsChanged: handleRowsEvent,
      rowsRemoved: handleRowsEvent,
      rowsReordered: handleRowsEvent,
      columnsAdded: handleColumnsEvent,
      columnsRemoved: handleColumnsEvent,
      columnsFlagsChanged: handleColumnsEvent
    }, function (handler, eventName) {
      var f = function (rowsOrColumns) {
        var args = _.toArray(arguments).slice(1);
        handler(rowsOrColumns, eventName, args);
      };
      tableModel.on(eventName, f);
      handlers[eventName] = f;
    });

    return Promise.all([
      updateThead(that, tableModel),
      updateTbody(that, tableModel)
    ]);
  };

  /**
   * Remove `TableWidget` class and event handlers from the loaded table model.
   */
  Table.prototype.doDestroy = function () {
    var that = this,
        jq = that.jq();

    jq.children('.tableContainer').off('scroll');

    that.$disarmModel();

    if (this.$getPaginationHandler()) {
      this.$getPaginationHandler().destroy();
    }

    this.$getSelection().removeAllListeners();

    return destroyCells(that, that.$getTbody())
      .then(function () {
        jq.removeClass('TableWidget fixedHeaders');
        return that.getChildWidgets().destroyAll();
      });
  };

  /**
   * Remove all listeners from previous table model before
   * arming the new one
   * @private
   *
   */
  Table.prototype.$disarmModel = function () {
    var handlers = this.$handlers,
        tableModel = this.value();

    //Remove all the model listeners
    if (handlers && tableModel) {
      _.each(handlers, function (handler, event) {
        tableModel.removeListener(event, handler);
      });
    }
  };

  /**
   * When showing a context menu, will decide which values in the TableModel are
   * the targets of the right-click operation.
   *
   * If the row being right-clicked is not already selected, then the subject of
   * the corresponding `Row` will be used to show the context menu.
   *
   * If the row being right-clicked is already selected, then the subjects of
   * *all* selected `Row`s will be used.
   *
   * @param {JQuery} elem
   * @returns {Array.<*>} array containing the subjects of the rows being
   * right-clicked. Can return an empty array if no rows are present.
   */
  Table.prototype.getSubject = function (elem) {
    var model = this.value();

    if (!model) {
      return [];
    }

    var index = elem.closest('tr').index(),
      selection = this.$getSelection(),
      rows = model.getRows();

    if (selection.isSelected(index)) {
      return _.map(selection.getSelectedElements(rows), function (selectedRow) {
        return selectedRow.getSubject();
      });
    }

    var row = rows[index];

    return row ? [row.getSubject()] : [];
  };

  /**
   * Get all rows which are currently selected by the user.
   * @returns {Array.<module:nmodule/webEditors/rc/wb/table/model/Row>}
   * @since Niagara 4.6
   */
  Table.prototype.getSelectedRows = function () {
    return this.$getSelection().getSelectedElements(this.value().getRows());
  };

  /**
   * Resolve display HTML for the given column and row. By default,
   * will simply proxy through to `Column#buildCell`.
   *
   * @private
   * @param {module:nmodule/webEditors/rc/wb/table/model/Column} column
   * @param {module:nmodule/webEditors/rc/wb/table/model/Row} row
   * @param {jQuery} dom the td element for the cell
   * @returns {Promise}
   */
  Table.prototype.buildCell = function (column, row, dom) {
    return this.$buildCell ? this.$buildCell(column, row, dom)
                           : column.buildCell(row, dom);
  };

  /**
   * Complementary function to `#buildCell`. This will give the model a chance
   * to clean up any resources allocated when creating the cell's HTML, such as
   * unhooking event handlers. As with cell construction, this will default to calling through
   * to `Column#destroyCell`.
   *
   * @private
   * @param {module:nmodule/webEditors/rc/wb/table/model/Column} column
   * @param {module:nmodule/webEditors/rc/wb/table/model/Row} row
   * @param {jQuery} dom the td element for the cell
   * @returns {Promise}
   */
  Table.prototype.destroyCell = function (column, row, dom) {
    return this.$destroyCell ? this.$destroyCell(column, row, dom)
                             : column.destroyCell(row, dom);
  };

  /**
   * Called when a row and its cells have been constructed. This will
   * allow any final dom customizations on the row with all its cells
   * constructed, before it is inserted into the table. By default, this
   * will not modify the dom.
   *
   * @private
   * @param {module:nmodule/webEditors/rc/wb/table/model/Column} row
   * @param {jQuery} dom the tr element constructed for the given `Row`.
   * @returns {Promise}
   */
  Table.prototype.finishBuildingRow = function (row, dom) {
    return this.$finishBuildingRow ? this.$finishBuildingRow(row, dom)
                                   : Promise.resolve(dom);
  };

  return Table;
});
