/*
 * @copyright 2016 Tridium, Inc. All Rights Reserved.
 * @author Vikram Nagulan
 */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/table/pagination/PaginationWidget
 */
define(['jquery',
        'bajaux/Widget',
        'hbs!nmodule/webEditors/rc/wb/template/Pagination',
        'nmodule/js/rc/tinyevents/tinyevents',
        'bajaux/commands/Command',
        'bajaux/util/CommandButtonGroup',
        'nmodule/webEditors/rc/fe/fe',
        'nmodule/webEditors/rc/fe/feDialogs',
        'lex!webEditors',
        'nmodule/webEditors/rc/util/storageUtil',
        'css!nmodule/webEditors/rc/fe/webEditors-structure'], function (
         $,
         Widget,
         tplPagination,
         tinyevents,
         Command,
         CommandButtonGroup,
         fe,
         feDialogs,
         lexs,
         storageUtil) {

  'use strict';

  var PAGE_CLASS = '.table-widget-page-number',
    NEXT_CLASS = '.table-widget-page-next',
    PREV_CLASS = '.table-widget-page-prev',
    MAX_PAGESIZE = 100,
    MIN_PAGESIZE = 10,
    lex = lexs[0],
    _sessionStorage;

  /**
   * Command to show the settings dialog.
   *
   * @inner
   * @class
   * @extends module:bajaux/commands/Command
   * @param {module:nmodule/webEditors/rc/wb/table/pagination/PaginationWidget} ed
   */
  var SettingsCommand = function SettingsCommand(ed) {
    Command.call(this, {
      module: 'webEditors',
      lex: "paginationWidget.commands.settings",
      func: function () {
        return feDialogs.showFor({
          value: ed.getPageSize(),
          properties: {
            max: MAX_PAGESIZE,
            prelabel: lex.get("paginationWidget.settings.dialog.label.pageSize"),
            precision: 0
          },
          title: lex.get("paginationWidget.settings.dialog.title")
        })
          .then(function (newpgSz) {
            if (newpgSz) {
              ed.properties().setValue("pageSize", newpgSz);
              ed.$storePageSize(newpgSz);
              ed.emit('pagechanged', 1);
            }
          });
      }
    });
  };
  SettingsCommand.prototype = Object.create(Command.prototype);
  SettingsCommand.prototype.constructor = SettingsCommand;

  /**
   * @private
   * @class
   * @extends module:bajaux/Widget
   * @alias module:nmodule/webEditors/rc/wb/table/pagination/PaginationWidget
   * @mixes tinyevents
   *
   * Pagination Widget provides a UX to set page limit and turn pages.
   * Note that this widget is not intended to be used as a "Widget" and
   * will be initialized only by a Table presently.
   * @see module:nmodule/webEditors/rc/wb/table/Table#doInitialize
   *
   * <p>
   * The widget emits a 'pageChanged' event when a page is changed and passes in the index
   * of the currently displayed/selected page. Table listens to this event and handles the
   * page load functionality.
   * </p>
   *
   */
  var PaginationWidget = function PaginationWidget() {
    var that = this;
    Widget.apply(this, arguments);
    tinyevents(this);
    //Default page size property
    that.properties().add({
      name: 'pageSize',
      value: that.$getStoredPageSize() || MIN_PAGESIZE,
      parameter: false,
      hidden: true,
      trans: false
    });
    that.getCommandGroup().add(
      new SettingsCommand(that)
    );
  };
  PaginationWidget.prototype = Object.create(Widget.prototype);
  PaginationWidget.prototype.constructor = PaginationWidget;

  /**
   * @param {JQuery} dom
   * @returns {Promise}
   */
  PaginationWidget.prototype.doInitialize = function (dom) {
    var that = this;
    dom.html(tplPagination({
      prevText: lex.get("paginationWidget.commands.prev"),
      nextText: lex.get("paginationWidget.commands.next"),
    }));

    // Page selection handler
    dom.on('change', PAGE_CLASS, function () {
      //Emit a pagechanged that a grid table can register to
      var pgIndex = parseInt($(this).val(), 10) || 0;
      that.emit('pagechanged', pgIndex < 0 ? 1 : pgIndex);
    });

    // Query text field handler (submit search on enter key press)
    dom.on('keypress', PAGE_CLASS, function (event) {
      if (event.which === 13) {
        var pgIndex = parseInt($(this).val(), 10) || 0;
        that.emit('pagechanged', pgIndex < 0 ? 1 : pgIndex);
        return false; //Prevent page from reloading
      }
    });

    // Next Page button handler
    dom.on('click', NEXT_CLASS, function () {
      var page = dom.find(PAGE_CLASS).first(),
        newPage = parseInt(page.val(), 10) + 1;
      page.val(newPage);
      page.change();
    });

    // Previous Page button handler
    dom.on('click', PREV_CLASS, function () {
      var page = dom.find(PAGE_CLASS).first(),
        newPage = parseInt(page.val(), 10) - 1;
      page.val(newPage);
      page.change();
    });

    return fe.buildFor({
      dom: $('.settings', that.jq()),
      type: CommandButtonGroup,
      value: this.getCommandGroup(),
      properties: {toolbar: true}
    });
  };

  PaginationWidget.prototype.getPageSize = function () {
    return parseInt(this.properties().getValue('pageSize'), 10);
  };

  /**
   * Update widget parameters (say after every page changed event)
   * @param {object} params - accepts
   * {
   *  isComplete:<boolean>, //false if more pages can be navigated back and forth
   *  currentIdx:<Number>, //The current selected page index
   *  noOfPages:<Number> //No of pages if available
   * }
   */
  PaginationWidget.prototype.updateWidgetDisplay = function (params) {
    var that = this;
    var nextCtrl = that.jq().find(NEXT_CLASS).first(),
      pageCtrl = that.jq().find(PAGE_CLASS).first(),
      prevCtrl = that.jq().find(PREV_CLASS).first();

    if (params) {
      var pgIndex = params.currentIdx || 1;
      var noOfPages = params.noOfPages || 0;
      var isComplete = pgIndex >= noOfPages;
      if (!isComplete) {
        nextCtrl.prop('disabled', false);
        pageCtrl.val(pgIndex);
      }
      else {
        //reset page index if it is out of bounds
        pgIndex = noOfPages;
        nextCtrl.prop('disabled', true);
        pageCtrl.val(noOfPages);
        pageCtrl.prop("max", noOfPages);
      }
      prevCtrl.prop('disabled', pgIndex <= 1);
      pageCtrl.prop('disabled', false);
    }
  };

  /**
   * @private
   * @returns {module:bajaux/commands/Command}
   */
  PaginationWidget.prototype.$getSettingsCommand = function () {
    return this.getCommandGroup().get(0);
  };

  /**
   * Clean up the dom and call Widget destroy
   * @returns {*}
   */
  PaginationWidget.prototype.doDestroy = function () {
    this.jq().find(".table-widget-page-commands").remove();
  };

  PaginationWidget.prototype.$getStoredPageSize = function () {
    return getSessionStorage().getItem('pageSize');
  };

  PaginationWidget.prototype.$storePageSize = function (pageSize) {
    getSessionStorage().setItem('pageSize', pageSize);
  };

  ////////////////Private functions/////////////////////////////////////

  /**
   * Utility to get session storage to save the page size
   * @returns {*|Object}
   */
  function getSessionStorage() {
    return _sessionStorage || (_sessionStorage = storageUtil.getSessionStorage());
  }

  return PaginationWidget;
});
