/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/util/UserPasswordEditor
 */
define(['baja!',
        'baja!baja:PasswordCache,' +
          'baja:PasswordAuthenticationScheme,baja:User',
        'jquery',
        'Promise',
        'nmodule/webEditors/rc/fe/baja/ConfirmPasswordEditor',
        'nmodule/webEditors/rc/fe/baja/BaseEditor',
        'nmodule/webEditors/rc/fe/baja/util/spaceUtils',
        'nmodule/webEditors/rc/fe/baja/util/subscriberUtils',
        'nmodule/webEditors/rc/servlets/password'], function (
         baja,
         types,
         $,
         Promise,
         ConfirmPasswordEditor,
         BaseEditor,
         spaceUtils,
         subscriberUtils,
         password) {

  'use strict';

  var resolveService = spaceUtils.resolveService,
      leaseWhenNeeded = subscriberUtils.leaseWhenNeeded,
      getAuthenticationSchemes = leaseWhenNeeded(function (base) {
        return resolveService('baja:AuthenticationService', base)
          .then(function (authenticationService) {
            return authenticationService.get('authenticationSchemes');
          });
      }),
      SHOULD_VALIDATE = BaseEditor.SHOULD_VALIDATE;

  /** @param {String} err */
  function reject(err) { return Promise.reject(new Error(err)); }

  function getPasswordConfig(user, base) {
    return getAuthenticationSchemes(base)
      .then(function (authenticationSchemes) {
        var schemeName = user.get('authenticationSchemeName');
        //NCCB-18386
        return baja.Ord.make({
          base: authenticationSchemes.getNavOrd(),
          child: 'slot:' + schemeName + '/globalPasswordConfiguration'
        })
          .get({ lease: true });
      });
  }


  /**
   * @private
   * @class
   * @alias module:nmodule/webEditors/rc/wb/util/UserPasswordEditor
   * @extends module:nmodule/webEditors/rc/fe/baja/BaseEditor
   */
  var UserPasswordEditor = function UserPasswordEditor(params) {
    var that = this;

    ConfirmPasswordEditor.call(that, $.extend({}, {
      keyName: 'UserPasswordEditor'
    }, params));

    /**
     * A UserPasswordEditor will fail to validate under the following conditions:
     *
     * - the editor is operating in an insecure environment - passwords must
     *   not be sent over the wire in cleartext
     * - the two entered passwords do not match
     * - the passwords do not meet the complexity requirements on the
     *   password authentication scheme configured for the backing User
     *
     * @method module:nmodule/webEditors/rc/fe/baja/UserPasswordEditor#validate
     */

    that.validators().add(function (val) {
      return that.$getPasswordConfig()
        .then(function (config) {
          var strength = config.get('passwordStrength');
          return strength.validatePassword(val.encodeToString())
            .catch(function (err) {
              //var arr = JSON.parse(err.message);
              var arr = err.message.split('\n');
              return that.$getUser().then( function(user) {
                throw new Error(user.getName() + ': ' + arr[0] + ' ' + arr.slice(1).join(', '));
              });
            });
        });
    });
  };
  UserPasswordEditor.prototype = Object.create(ConfirmPasswordEditor.prototype);
  UserPasswordEditor.prototype.constructor = UserPasswordEditor;

  /**
   * Get the `baja:User` the loaded password belongs to.
   *
   * @private
   * @returns {Promise} promise to be resolved with a `User`, or
   * rejected if the password does not belong to a
   * `baja:PasswordCache` that belongs to a `baja:User`.
   */
  UserPasswordEditor.prototype.$getUser = function () {
    var complex = this.getComplex();

    if (!complex) {
      return reject('UserPasswordEditor only works as a ComplexSlotEditor');
    }

    if (!baja.hasType(complex, 'baja:PasswordCache')) {
      return reject('password must belong to a PasswordCache');
    }

    var user = complex.getParent();

    if (!baja.hasType(user, 'baja:User')) {
      return reject('authenticator must belong to a User');
    }
    return Promise.resolve(user);
  };

  /**
   * Resolve the `baja:GlobalPasswordConfiguration` that applies to the backing
   * `User`, as configured by the `authenticationSchemeName` slot on the user.
   *
   * @private
   * @returns {Promise} promise to be resolved with the password config,
   * or rejected if it could not be found
   */
  UserPasswordEditor.prototype.$getPasswordConfig = function () {
    return Promise.join(this.$getUser(), this.getOrdBase())
      .spread(getPasswordConfig);
  };

  /**
   * Does not actually load the password into the DOM. This only verifies that
   * the loaded password belongs to a `baja:User`; this means that in order
   * to function, this `UserPasswordEditor` must be instantiated as a
   * `ComplexSlotEditor`. The backing `Complex` instance *must* be a
   * `baja:PasswordCache` or subclass that belongs to a
   * `baja:User`.
   *
   * @param {baja.Simple} password
   * @returns {Promise} promise to be resolved if the password actually
   * is a `baja:Password` that correctly belongs to a `User`.
   */
  UserPasswordEditor.prototype.doLoad = function (password) {
    var that = this;
    if (!baja.hasType(password, 'baja:Password')) {
      return reject('baja:Password required');
    }
    return Promise.join(this.$getUser(), ConfirmPasswordEditor.prototype.doLoad.call(this, password))
      .spread(function (user) {
        /**
         * when adding a new user, we want the password to always be validated.
         * when editing an existing user, we only want to validate the password
         * if it has been modified. since new users are not mounted yet, we set
         * shouldValidate accordingly, which will be respected by UserManager.
         */
        that.properties().add(SHOULD_VALIDATE, !user.isMounted());
      });
  };

  /**
   * Saving a `UserPasswordEditor` only makes sense when editing a password that
   * belongs to a User. Since BajaScript will not encode a `baja:Password`
   * (nor should it), the save operation will be delegated to the
   * `ChangeUserPasswordServlet` instead.
   *
   * @see module:nmodule/webEditors/rc/fe/baja/ComplexSlotEditor
   * @see module:nmodule/webEditors/rc/servlets/password
   * @returns {Promise}
   */
  UserPasswordEditor.prototype.saveToComplex = function (pw, params) {
    return Promise.join(this.$getUser(), this.getOrdBase())
      .spread(function (user, base) {
        return password.changePasswords(
          [user.getName()],
          [pw.encodeToString()],
          base
        );
      });
  };

  return (UserPasswordEditor);
});
