/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Logan Byam
 */

/**
 * @private
 * @module nmodule/webEditors/rc/wb/util/user/UserManagerMgrModel
 */
define(['baja!',
        'baja!baja:PasswordAuthenticator,baja:User',
        'Promise',
        'underscore',
        'nmodule/webEditors/rc/servlets/password',
        'nmodule/webEditors/rc/wb/mgr/MgrTypeInfo',
        'nmodule/webEditors/rc/wb/mgr/model/MgrModel',
        'nmodule/webEditors/rc/wb/mgr/model/mgrAgents',
        'nmodule/webEditors/rc/wb/mgr/model/columns/NameMgrColumn',
        'nmodule/webEditors/rc/wb/mgr/model/columns/PropertyMgrColumn',
        'nmodule/webEditors/rc/wb/table/model/Column',
        'nmodule/webEditors/rc/wb/table/model/source/ContainerComponentSource',
        'nmodule/webEditors/rc/wb/util/user/AuthenticatorMgrColumn',
        'nmodule/webEditors/rc/wb/util/user/ExpirationMgrColumn'], function (
         baja,
         types,
         Promise,
         _,
         password,
         MgrTypeInfo,
         MgrModel,
         mgrAgents,
         NameMgrColumn,
         PropertyMgrColumn,
         Column,
         ContainerComponentSource,
         AuthenticatorMgrColumn,
         ExpirationMgrColumn) {

  'use strict';

  var USER_TYPE = baja.lt('baja:User'),
    EDITABLE = Column.flags.EDITABLE,
    UNSEEN = Column.flags.UNSEEN,
    defaultPassword = baja.$('baja:PasswordAuthenticator').get('password');

////////////////////////////////////////////////////////////////
// Column definitions
////////////////////////////////////////////////////////////////

  function toColumn(slot, params) {
    params = params || {};
    var Ctor = params.ctor || PropertyMgrColumn,
      flags = params.flags || EDITABLE;

    return new Ctor(slot, { type: USER_TYPE, flags: flags });
  }

  var COLUMNS = [
    new NameMgrColumn({ flags: EDITABLE }),
    toColumn('fullName'),
    toColumn('enabled'),
    toColumn('expiration', { ctor: ExpirationMgrColumn }),
    toColumn('lockOut', { flags: UNSEEN }),
    toColumn('roles', { flags: EDITABLE | UNSEEN }),
    toColumn('allowConcurrentSessions'),
    toColumn('autoLogoffSettings'),
    toColumn('networkUser'),
    toColumn('prototypeName'),
    toColumn('language'),
    toColumn('authenticationSchemeName'),
    toColumn('authenticator', { ctor: AuthenticatorMgrColumn, flags: EDITABLE | UNSEEN }),
    toColumn('email'),
    toColumn('cellPhoneNumber', { flags: EDITABLE | UNSEEN }),
    toColumn('facets', { flags: EDITABLE | UNSEEN }),
    toColumn('navFile', { flags: EDITABLE | UNSEEN })
  ];

  /**
   * Get component name.
   *
   * @param {baja.Complex} comp
   * @returns {String}
   */
  function toName(comp) { return comp.getName(); }

  /**
   * Get the user's password as a string, then reset the password to the
   * default so BajaScript will not attempt to encode it and send it over the
   * wire.
   *
   * @param {baja.Complex} user
   * @returns {String}
   */
  function extractPassword(user) {
    var authenticator = user.get('authenticator'),
        pw = authenticator.get('password');

    //if no password entered, treat it as blank.
    if (pw.equals(defaultPassword)) {
      pw = baja.$('baja:Password');
    }

    authenticator.set({
      slot: 'password',
      value: defaultPassword
    });

    return pw.encodeToString();
  }

  /**
   * Get any passwords on the user's authenticator in addition to the 'password'
   * property and set them over RPC with the provided batch, then reset the password
   * to the default so BajaScript will not attempt to encode it and send it over the wire.
   *
   * @param {baja.Complex} user
   * @param {baja.Component} userService
   * @param {baja.comm.Batch} batch
   * @returns {Promise}
   */
  function setExtraPasswords(user, userService, batch) {
    var authenticator = user.get('authenticator'),
      promises = [],
      authenticatorOrd = baja.Ord.make(userService.getNavOrd() +  '/' + toName(user) + '/authenticator');

    authenticator.getSlots().properties().is("baja:Password").each(function (slot) {
      if (slot.getName() !== 'password') {
        var pw = authenticator.get(slot);
        if (!pw.equals(defaultPassword)) {
          promises.push(password.setPassword(pw.encodeToString(), slot.getName(), authenticatorOrd, {batch: batch}));
        }

        authenticator.set({
          slot: slot,
          value: defaultPassword
        });
      }
    });

    return Promise.all(promises);
  }

  /**
   * `MgrModel` for the User Manager. Don't instantiate directly; use the `make`
   * method.
   *
   * @class
   * @private
   * @extends module:nmodule/webEditors/rc/wb/mgr/model/MgrModel
   * @alias module:nmodule/webEditors/rc/wb/util/user/UserManagerMgrModel
   */
  var UserManagerMgrModel = function UserManagerMgrModel() {
    MgrModel.apply(this, arguments);
  };
  UserManagerMgrModel.prototype = Object.create(MgrModel.prototype);
  UserManagerMgrModel.prototype.constructor = UserManagerMgrModel;

  /**
   * Create a `UserManagerMgrModel` instance with the appropriate columns.
   *
   * @param {baja.Component} userService
   * @returns {Promise.<module:nmodule/webEditors/rc/wb/util/user/UserManagerMgrModel>}
   */
  UserManagerMgrModel.make = function (userService) {
    return Promise.all([
      MgrTypeInfo.make([ USER_TYPE ]),
      mgrAgents.getMixInColumns(USER_TYPE)
    ])
      .spread(function (newTypes, mixinCols) {
        return new UserManagerMgrModel({
          columns: COLUMNS.concat(mixinCols),
          componentSource: new ContainerComponentSource({
            container: userService,
            filter: [ 'baja:User' ]
          }),
          newTypes: newTypes
        });
      });
  };

  /**
   * When creating a new user, copy the default prototype off of
   * the user service. See `wbutil:UserManager$model#newInstance`.
   *
   * @param {module:nmodule/webEditors/rc/wb/mgr/MgrTypeInfo} typeInfo
   * @returns {baja.Component} a `baja:User` cloned from the default
   * prototype.
   */
  UserManagerMgrModel.prototype.newInstance = function (typeInfo) {
    //user manager always uses a UserService as component container.
    var componentSource = this.getComponentSource(),
        userService = componentSource.getContainer(),
        prototypes = userService.get('userPrototypes');

    // noinspection JSValidateTypes
    return prototypes.get('defaultPrototype').newCopy();
  };

  /**
   * Before writing the users up to the station, wipe all the passwords off
   * (BajaScript will not encode them). After the users are added, then send
   * the passwords up through the designated `ChangeUserPasswordServlet`.
   *
   * @param {Array.<baja.Component>} users
   * @returns {Promise}
   */
  UserManagerMgrModel.prototype.addInstances = function (users) {
    var names = _.map(users, toName),
        passwords = _.map(users, extractPassword),
        userService = this.getComponentSource().getContainer(),
        batch = new baja.comm.Batch(),
        extraPasswordPromise = Promise.all(_.map(users, function(user) {
          return setExtraPasswords(user, userService, batch);
        }));

    return MgrModel.prototype.addInstances.apply(this, arguments)
      .then(function () {
        var promise = Promise.join(password.changePasswords(names, passwords, userService, {batch: batch}), extraPasswordPromise);
        return batch.commit(promise);
      });
  };

  return (UserManagerMgrModel);
});
